from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.custom_authentication_type import CustomAuthenticationType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.custom_integration_extra_type_0 import CustomIntegrationExtraType0


T = TypeVar("T", bound="CustomIntegration")


@_attrs_define
class CustomIntegration:
    """
    Attributes
    ----------
        models (list[str]): List of model names for the custom integration
        endpoint (str): Endpoint URL for the custom integration.
        authentication_type (Union[Unset, CustomAuthenticationType]):
        default_model (Union[None, Unset, str]): Default model to use. If not provided, defaults to the first model in
            the models list.
        authentication_scope (Union[None, Unset, str]): Optional scope for OAuth2 authentication.
        oauth2_token_url (Union[None, Unset, str]): OAuth2 token URL for custom OAuth2 authentication. If not provided,
            defaults to the endpoint.
        id (Union[None, Unset, str]):
        name (Union[Literal['custom'], Unset]):  Default: 'custom'.
        extra (Union['CustomIntegrationExtraType0', None, Unset]):
    """

    models: list[str]
    endpoint: str
    authentication_type: Union[Unset, CustomAuthenticationType] = UNSET
    default_model: Union[None, Unset, str] = UNSET
    authentication_scope: Union[None, Unset, str] = UNSET
    oauth2_token_url: Union[None, Unset, str] = UNSET
    id: Union[None, Unset, str] = UNSET
    name: Union[Literal["custom"], Unset] = "custom"
    extra: Union["CustomIntegrationExtraType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.custom_integration_extra_type_0 import CustomIntegrationExtraType0

        models = self.models

        endpoint = self.endpoint

        authentication_type: Union[Unset, str] = UNSET
        if not isinstance(self.authentication_type, Unset):
            authentication_type = self.authentication_type.value

        default_model: Union[None, Unset, str]
        default_model = UNSET if isinstance(self.default_model, Unset) else self.default_model

        authentication_scope: Union[None, Unset, str]
        authentication_scope = UNSET if isinstance(self.authentication_scope, Unset) else self.authentication_scope

        oauth2_token_url: Union[None, Unset, str]
        oauth2_token_url = UNSET if isinstance(self.oauth2_token_url, Unset) else self.oauth2_token_url

        id: Union[None, Unset, str]
        id = UNSET if isinstance(self.id, Unset) else self.id

        name = self.name

        extra: Union[None, Unset, dict[str, Any]]
        if isinstance(self.extra, Unset):
            extra = UNSET
        elif isinstance(self.extra, CustomIntegrationExtraType0):
            extra = self.extra.to_dict()
        else:
            extra = self.extra

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"models": models, "endpoint": endpoint})
        if authentication_type is not UNSET:
            field_dict["authentication_type"] = authentication_type
        if default_model is not UNSET:
            field_dict["default_model"] = default_model
        if authentication_scope is not UNSET:
            field_dict["authentication_scope"] = authentication_scope
        if oauth2_token_url is not UNSET:
            field_dict["oauth2_token_url"] = oauth2_token_url
        if id is not UNSET:
            field_dict["id"] = id
        if name is not UNSET:
            field_dict["name"] = name
        if extra is not UNSET:
            field_dict["extra"] = extra

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.custom_integration_extra_type_0 import CustomIntegrationExtraType0

        d = dict(src_dict)
        models = cast(list[str], d.pop("models"))

        endpoint = d.pop("endpoint")

        _authentication_type = d.pop("authentication_type", UNSET)
        authentication_type: Union[Unset, CustomAuthenticationType]
        if isinstance(_authentication_type, Unset):
            authentication_type = UNSET
        else:
            authentication_type = CustomAuthenticationType(_authentication_type)

        def _parse_default_model(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        default_model = _parse_default_model(d.pop("default_model", UNSET))

        def _parse_authentication_scope(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        authentication_scope = _parse_authentication_scope(d.pop("authentication_scope", UNSET))

        def _parse_oauth2_token_url(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        oauth2_token_url = _parse_oauth2_token_url(d.pop("oauth2_token_url", UNSET))

        def _parse_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        id = _parse_id(d.pop("id", UNSET))

        name = cast(Union[Literal["custom"], Unset], d.pop("name", UNSET))
        if name != "custom" and not isinstance(name, Unset):
            raise ValueError(f"name must match const 'custom', got '{name}'")

        def _parse_extra(data: object) -> Union["CustomIntegrationExtraType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return CustomIntegrationExtraType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["CustomIntegrationExtraType0", None, Unset], data)

        extra = _parse_extra(d.pop("extra", UNSET))

        custom_integration = cls(
            models=models,
            endpoint=endpoint,
            authentication_type=authentication_type,
            default_model=default_model,
            authentication_scope=authentication_scope,
            oauth2_token_url=oauth2_token_url,
            id=id,
            name=name,
            extra=extra,
        )

        custom_integration.additional_properties = d
        return custom_integration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ExperimentDatasetRequest")


@_attrs_define
class ExperimentDatasetRequest:
    """
    Attributes
    ----------
        dataset_id (str):
        version_index (int):
    """

    dataset_id: str
    version_index: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        dataset_id = self.dataset_id

        version_index = self.version_index

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"dataset_id": dataset_id, "version_index": version_index})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        dataset_id = d.pop("dataset_id")

        version_index = d.pop("version_index")

        experiment_dataset_request = cls(dataset_id=dataset_id, version_index=version_index)

        experiment_dataset_request.additional_properties = d
        return experiment_dataset_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Name")


@_attrs_define
class Name:
    """Global name class for handling unique naming across the application.

    Attributes
    ----------
        value (str):
        append_suffix_if_duplicate (Union[Unset, bool]):  Default: False.
    """

    value: str
    append_suffix_if_duplicate: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        value = self.value

        append_suffix_if_duplicate = self.append_suffix_if_duplicate

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"value": value})
        if append_suffix_if_duplicate is not UNSET:
            field_dict["append_suffix_if_duplicate"] = append_suffix_if_duplicate

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        value = d.pop("value")

        append_suffix_if_duplicate = d.pop("append_suffix_if_duplicate", UNSET)

        name = cls(value=value, append_suffix_if_duplicate=append_suffix_if_duplicate)

        name.additional_properties = d
        return name

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Metrics")


@_attrs_define
class Metrics:
    """
    Attributes
    ----------
        duration_ns (Union[None, Unset, int]): Duration of the trace or span in nanoseconds.  Displayed as 'Latency' in
            Galileo.
    """

    duration_ns: Union[None, Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        duration_ns: Union[None, Unset, int]
        duration_ns = UNSET if isinstance(self.duration_ns, Unset) else self.duration_ns

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if duration_ns is not UNSET:
            field_dict["duration_ns"] = duration_ns

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_duration_ns(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        duration_ns = _parse_duration_ns(d.pop("duration_ns", UNSET))

        metrics = cls(duration_ns=duration_ns)

        metrics.additional_properties = d
        return metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class RunCreatedAtFilterOperator(str, Enum):
    EQ = "eq"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    NE = "ne"

    def __str__(self) -> str:
        return str(self.value)



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.image_generation_event import ImageGenerationEvent
    from ..models.internal_tool_call import InternalToolCall
    from ..models.llm_metrics import LlmMetrics
    from ..models.llm_span_dataset_metadata import LlmSpanDatasetMetadata
    from ..models.llm_span_tools_type_0_item import LlmSpanToolsType0Item
    from ..models.llm_span_user_metadata import LlmSpanUserMetadata
    from ..models.mcp_approval_request_event import MCPApprovalRequestEvent
    from ..models.mcp_call_event import MCPCallEvent
    from ..models.mcp_list_tools_event import MCPListToolsEvent
    from ..models.message import Message
    from ..models.message_event import MessageEvent
    from ..models.reasoning_event import ReasoningEvent


T = TypeVar("T", bound="LlmSpan")


@_attrs_define
class LlmSpan:
    """
    Attributes
    ----------
        type_ (Union[Literal['llm'], Unset]): Type of the trace, span or session. Default: 'llm'.
        input_ (Union[Unset, list['Message']]): Input to the trace or span.
        redacted_input (Union[None, Unset, list['Message']]): Redacted input of the trace or span.
        output (Union[Unset, Message]):
        redacted_output (Union['Message', None, Unset]): Redacted output of the trace or span.
        name (Union[Unset, str]): Name of the trace, span or session. Default: ''.
        created_at (Union[Unset, datetime.datetime]): Timestamp of the trace or span's creation.
        user_metadata (Union[Unset, LlmSpanUserMetadata]): Metadata associated with this trace or span.
        tags (Union[Unset, list[str]]): Tags associated with this trace or span.
        status_code (Union[None, Unset, int]): Status code of the trace or span. Used for logging failure or error
            states.
        metrics (Union[Unset, LlmMetrics]):
        external_id (Union[None, Unset, str]): A user-provided session, trace or span ID.
        dataset_input (Union[None, Unset, str]): Input to the dataset associated with this trace
        dataset_output (Union[None, Unset, str]): Output from the dataset associated with this trace
        dataset_metadata (Union[Unset, LlmSpanDatasetMetadata]): Metadata from the dataset associated with this trace
        id (Union[None, Unset, str]): Galileo ID of the session, trace or span
        session_id (Union[None, Unset, str]): Galileo ID of the session containing the trace or span or session
        trace_id (Union[None, Unset, str]): Galileo ID of the trace containing the span (or the same value as id for a
            trace)
        step_number (Union[None, Unset, int]): Topological step number of the span.
        parent_id (Union[None, Unset, str]): Galileo ID of the parent of this span
        tools (Union[None, Unset, list['LlmSpanToolsType0Item']]): List of available tools passed to the LLM on
            invocation.
        events (Union[None, Unset, list[Union['ImageGenerationEvent', 'InternalToolCall', 'MCPApprovalRequestEvent',
            'MCPCallEvent', 'MCPListToolsEvent', 'MessageEvent', 'ReasoningEvent']]]): List of reasoning, internal tool
            call, or MCP events that occurred during the LLM span.
        model (Union[None, Unset, str]): Model used for this span.
        temperature (Union[None, Unset, float]): Temperature used for generation.
        finish_reason (Union[None, Unset, str]): Reason for finishing.
    """

    type_: Union[Literal["llm"], Unset] = "llm"
    input_: Union[Unset, list["Message"]] = UNSET
    redacted_input: Union[None, Unset, list["Message"]] = UNSET
    output: Union[Unset, "Message"] = UNSET
    redacted_output: Union["Message", None, Unset] = UNSET
    name: Union[Unset, str] = ""
    created_at: Union[Unset, datetime.datetime] = UNSET
    user_metadata: Union[Unset, "LlmSpanUserMetadata"] = UNSET
    tags: Union[Unset, list[str]] = UNSET
    status_code: Union[None, Unset, int] = UNSET
    metrics: Union[Unset, "LlmMetrics"] = UNSET
    external_id: Union[None, Unset, str] = UNSET
    dataset_input: Union[None, Unset, str] = UNSET
    dataset_output: Union[None, Unset, str] = UNSET
    dataset_metadata: Union[Unset, "LlmSpanDatasetMetadata"] = UNSET
    id: Union[None, Unset, str] = UNSET
    session_id: Union[None, Unset, str] = UNSET
    trace_id: Union[None, Unset, str] = UNSET
    step_number: Union[None, Unset, int] = UNSET
    parent_id: Union[None, Unset, str] = UNSET
    tools: Union[None, Unset, list["LlmSpanToolsType0Item"]] = UNSET
    events: Union[
        None,
        Unset,
        list[
            Union[
                "ImageGenerationEvent",
                "InternalToolCall",
                "MCPApprovalRequestEvent",
                "MCPCallEvent",
                "MCPListToolsEvent",
                "MessageEvent",
                "ReasoningEvent",
            ]
        ],
    ] = UNSET
    model: Union[None, Unset, str] = UNSET
    temperature: Union[None, Unset, float] = UNSET
    finish_reason: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.image_generation_event import ImageGenerationEvent
        from ..models.internal_tool_call import InternalToolCall
        from ..models.mcp_call_event import MCPCallEvent
        from ..models.mcp_list_tools_event import MCPListToolsEvent
        from ..models.message import Message
        from ..models.message_event import MessageEvent
        from ..models.reasoning_event import ReasoningEvent

        type_ = self.type_

        input_: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.input_, Unset):
            input_ = []
            for input_item_data in self.input_:
                input_item = input_item_data.to_dict()
                input_.append(input_item)

        redacted_input: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.redacted_input, Unset):
            redacted_input = UNSET
        elif isinstance(self.redacted_input, list):
            redacted_input = []
            for redacted_input_type_0_item_data in self.redacted_input:
                redacted_input_type_0_item = redacted_input_type_0_item_data.to_dict()
                redacted_input.append(redacted_input_type_0_item)

        else:
            redacted_input = self.redacted_input

        output: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.output, Unset):
            output = self.output.to_dict()

        redacted_output: Union[None, Unset, dict[str, Any]]
        if isinstance(self.redacted_output, Unset):
            redacted_output = UNSET
        elif isinstance(self.redacted_output, Message):
            redacted_output = self.redacted_output.to_dict()
        else:
            redacted_output = self.redacted_output

        name = self.name

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        user_metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.user_metadata, Unset):
            user_metadata = self.user_metadata.to_dict()

        tags: Union[Unset, list[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        status_code: Union[None, Unset, int]
        status_code = UNSET if isinstance(self.status_code, Unset) else self.status_code

        metrics: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.metrics, Unset):
            metrics = self.metrics.to_dict()

        external_id: Union[None, Unset, str]
        external_id = UNSET if isinstance(self.external_id, Unset) else self.external_id

        dataset_input: Union[None, Unset, str]
        dataset_input = UNSET if isinstance(self.dataset_input, Unset) else self.dataset_input

        dataset_output: Union[None, Unset, str]
        dataset_output = UNSET if isinstance(self.dataset_output, Unset) else self.dataset_output

        dataset_metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.dataset_metadata, Unset):
            dataset_metadata = self.dataset_metadata.to_dict()

        id: Union[None, Unset, str]
        id = UNSET if isinstance(self.id, Unset) else self.id

        session_id: Union[None, Unset, str]
        session_id = UNSET if isinstance(self.session_id, Unset) else self.session_id

        trace_id: Union[None, Unset, str]
        trace_id = UNSET if isinstance(self.trace_id, Unset) else self.trace_id

        step_number: Union[None, Unset, int]
        step_number = UNSET if isinstance(self.step_number, Unset) else self.step_number

        parent_id: Union[None, Unset, str]
        parent_id = UNSET if isinstance(self.parent_id, Unset) else self.parent_id

        tools: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.tools, Unset):
            tools = UNSET
        elif isinstance(self.tools, list):
            tools = []
            for tools_type_0_item_data in self.tools:
                tools_type_0_item = tools_type_0_item_data.to_dict()
                tools.append(tools_type_0_item)

        else:
            tools = self.tools

        events: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.events, Unset):
            events = UNSET
        elif isinstance(self.events, list):
            events = []
            for events_type_0_item_data in self.events:
                events_type_0_item: dict[str, Any]
                if isinstance(
                    events_type_0_item_data,
                    (
                        MessageEvent,
                        ReasoningEvent,
                        InternalToolCall,
                        ImageGenerationEvent,
                        MCPCallEvent,
                        MCPListToolsEvent,
                    ),
                ):
                    events_type_0_item = events_type_0_item_data.to_dict()
                else:
                    events_type_0_item = events_type_0_item_data.to_dict()

                events.append(events_type_0_item)

        else:
            events = self.events

        model: Union[None, Unset, str]
        model = UNSET if isinstance(self.model, Unset) else self.model

        temperature: Union[None, Unset, float]
        temperature = UNSET if isinstance(self.temperature, Unset) else self.temperature

        finish_reason: Union[None, Unset, str]
        finish_reason = UNSET if isinstance(self.finish_reason, Unset) else self.finish_reason

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if type_ is not UNSET:
            field_dict["type"] = type_
        if input_ is not UNSET:
            field_dict["input"] = input_
        if redacted_input is not UNSET:
            field_dict["redacted_input"] = redacted_input
        if output is not UNSET:
            field_dict["output"] = output
        if redacted_output is not UNSET:
            field_dict["redacted_output"] = redacted_output
        if name is not UNSET:
            field_dict["name"] = name
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if user_metadata is not UNSET:
            field_dict["user_metadata"] = user_metadata
        if tags is not UNSET:
            field_dict["tags"] = tags
        if status_code is not UNSET:
            field_dict["status_code"] = status_code
        if metrics is not UNSET:
            field_dict["metrics"] = metrics
        if external_id is not UNSET:
            field_dict["external_id"] = external_id
        if dataset_input is not UNSET:
            field_dict["dataset_input"] = dataset_input
        if dataset_output is not UNSET:
            field_dict["dataset_output"] = dataset_output
        if dataset_metadata is not UNSET:
            field_dict["dataset_metadata"] = dataset_metadata
        if id is not UNSET:
            field_dict["id"] = id
        if session_id is not UNSET:
            field_dict["session_id"] = session_id
        if trace_id is not UNSET:
            field_dict["trace_id"] = trace_id
        if step_number is not UNSET:
            field_dict["step_number"] = step_number
        if parent_id is not UNSET:
            field_dict["parent_id"] = parent_id
        if tools is not UNSET:
            field_dict["tools"] = tools
        if events is not UNSET:
            field_dict["events"] = events
        if model is not UNSET:
            field_dict["model"] = model
        if temperature is not UNSET:
            field_dict["temperature"] = temperature
        if finish_reason is not UNSET:
            field_dict["finish_reason"] = finish_reason

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.image_generation_event import ImageGenerationEvent
        from ..models.internal_tool_call import InternalToolCall
        from ..models.llm_metrics import LlmMetrics
        from ..models.llm_span_dataset_metadata import LlmSpanDatasetMetadata
        from ..models.llm_span_tools_type_0_item import LlmSpanToolsType0Item
        from ..models.llm_span_user_metadata import LlmSpanUserMetadata
        from ..models.mcp_approval_request_event import MCPApprovalRequestEvent
        from ..models.mcp_call_event import MCPCallEvent
        from ..models.mcp_list_tools_event import MCPListToolsEvent
        from ..models.message import Message
        from ..models.message_event import MessageEvent
        from ..models.reasoning_event import ReasoningEvent

        d = dict(src_dict)
        type_ = cast(Union[Literal["llm"], Unset], d.pop("type", UNSET))
        if type_ != "llm" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'llm', got '{type_}'")

        input_ = []
        _input_ = d.pop("input", UNSET)
        for input_item_data in _input_ or []:
            input_item = Message.from_dict(input_item_data)

            input_.append(input_item)

        def _parse_redacted_input(data: object) -> Union[None, Unset, list["Message"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                redacted_input_type_0 = []
                _redacted_input_type_0 = data
                for redacted_input_type_0_item_data in _redacted_input_type_0:
                    redacted_input_type_0_item = Message.from_dict(redacted_input_type_0_item_data)

                    redacted_input_type_0.append(redacted_input_type_0_item)

                return redacted_input_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["Message"]], data)

        redacted_input = _parse_redacted_input(d.pop("redacted_input", UNSET))

        _output = d.pop("output", UNSET)
        output: Union[Unset, Message]
        output = UNSET if isinstance(_output, Unset) else Message.from_dict(_output)

        def _parse_redacted_output(data: object) -> Union["Message", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return Message.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["Message", None, Unset], data)

        redacted_output = _parse_redacted_output(d.pop("redacted_output", UNSET))

        name = d.pop("name", UNSET)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        created_at = UNSET if isinstance(_created_at, Unset) else isoparse(_created_at)

        _user_metadata = d.pop("user_metadata", UNSET)
        user_metadata: Union[Unset, LlmSpanUserMetadata]
        user_metadata = UNSET if isinstance(_user_metadata, Unset) else LlmSpanUserMetadata.from_dict(_user_metadata)

        tags = cast(list[str], d.pop("tags", UNSET))

        def _parse_status_code(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        status_code = _parse_status_code(d.pop("status_code", UNSET))

        _metrics = d.pop("metrics", UNSET)
        metrics: Union[Unset, LlmMetrics]
        metrics = UNSET if isinstance(_metrics, Unset) else LlmMetrics.from_dict(_metrics)

        def _parse_external_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        external_id = _parse_external_id(d.pop("external_id", UNSET))

        def _parse_dataset_input(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_input = _parse_dataset_input(d.pop("dataset_input", UNSET))

        def _parse_dataset_output(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_output = _parse_dataset_output(d.pop("dataset_output", UNSET))

        _dataset_metadata = d.pop("dataset_metadata", UNSET)
        dataset_metadata: Union[Unset, LlmSpanDatasetMetadata]
        if isinstance(_dataset_metadata, Unset):
            dataset_metadata = UNSET
        else:
            dataset_metadata = LlmSpanDatasetMetadata.from_dict(_dataset_metadata)

        def _parse_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        id = _parse_id(d.pop("id", UNSET))

        def _parse_session_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        session_id = _parse_session_id(d.pop("session_id", UNSET))

        def _parse_trace_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        trace_id = _parse_trace_id(d.pop("trace_id", UNSET))

        def _parse_step_number(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        step_number = _parse_step_number(d.pop("step_number", UNSET))

        def _parse_parent_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        parent_id = _parse_parent_id(d.pop("parent_id", UNSET))

        def _parse_tools(data: object) -> Union[None, Unset, list["LlmSpanToolsType0Item"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                tools_type_0 = []
                _tools_type_0 = data
                for tools_type_0_item_data in _tools_type_0:
                    tools_type_0_item = LlmSpanToolsType0Item.from_dict(tools_type_0_item_data)

                    tools_type_0.append(tools_type_0_item)

                return tools_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["LlmSpanToolsType0Item"]], data)

        tools = _parse_tools(d.pop("tools", UNSET))

        def _parse_events(
            data: object,
        ) -> Union[
            None,
            Unset,
            list[
                Union[
                    "ImageGenerationEvent",
                    "InternalToolCall",
                    "MCPApprovalRequestEvent",
                    "MCPCallEvent",
                    "MCPListToolsEvent",
                    "MessageEvent",
                    "ReasoningEvent",
                ]
            ],
        ]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                events_type_0 = []
                _events_type_0 = data
                for events_type_0_item_data in _events_type_0:

                    def _parse_events_type_0_item(
                        data: object,
                    ) -> Union[
                        "ImageGenerationEvent",
                        "InternalToolCall",
                        "MCPApprovalRequestEvent",
                        "MCPCallEvent",
                        "MCPListToolsEvent",
                        "MessageEvent",
                        "ReasoningEvent",
                    ]:
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return MessageEvent.from_dict(data)

                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return ReasoningEvent.from_dict(data)

                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return InternalToolCall.from_dict(data)

                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return ImageGenerationEvent.from_dict(data)

                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return MCPCallEvent.from_dict(data)

                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return MCPListToolsEvent.from_dict(data)

                        except:  # noqa: E722
                            pass
                        if not isinstance(data, dict):
                            raise TypeError()
                        return MCPApprovalRequestEvent.from_dict(data)

                    events_type_0_item = _parse_events_type_0_item(events_type_0_item_data)

                    events_type_0.append(events_type_0_item)

                return events_type_0
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    None,
                    Unset,
                    list[
                        Union[
                            "ImageGenerationEvent",
                            "InternalToolCall",
                            "MCPApprovalRequestEvent",
                            "MCPCallEvent",
                            "MCPListToolsEvent",
                            "MessageEvent",
                            "ReasoningEvent",
                        ]
                    ],
                ],
                data,
            )

        events = _parse_events(d.pop("events", UNSET))

        def _parse_model(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        model = _parse_model(d.pop("model", UNSET))

        def _parse_temperature(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        temperature = _parse_temperature(d.pop("temperature", UNSET))

        def _parse_finish_reason(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        finish_reason = _parse_finish_reason(d.pop("finish_reason", UNSET))

        llm_span = cls(
            type_=type_,
            input_=input_,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            metrics=metrics,
            external_id=external_id,
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata,
            id=id,
            session_id=session_id,
            trace_id=trace_id,
            step_number=step_number,
            parent_id=parent_id,
            tools=tools,
            events=events,
            model=model,
            temperature=temperature,
            finish_reason=finish_reason,
        )

        llm_span.additional_properties = d
        return llm_span

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import contextlib
import random
from time import sleep

from pydantic import UUID4
from tqdm.auto import tqdm

from galileo.config import GalileoPythonConfig
from galileo.resources.api.jobs import (
    get_job_jobs_job_id_get,
    get_jobs_for_project_run_projects_project_id_runs_run_id_jobs_get,
)
from galileo.resources.models import HTTPValidationError, JobDB
from galileo.utils.logging import get_logger
from galileo_core.constants.job import JobName, JobStatus
from galileo_core.constants.scorers import Scorers

_logger = get_logger(__name__)


def get_job(job_id: str) -> JobDB:
    config = GalileoPythonConfig.get()

    response = get_job_jobs_job_id_get.sync(client=config.api_client, job_id=str(job_id))

    if isinstance(response, HTTPValidationError):
        raise ValueError(response.detail)
    if not response:
        raise ValueError(f"Failed to get job status for job {job_id}")
    return response


def get_run_scorer_jobs(project_id: str, run_id: str) -> list[JobDB]:
    config = GalileoPythonConfig.get()

    response = get_jobs_for_project_run_projects_project_id_runs_run_id_jobs_get.sync(
        client=config.api_client, project_id=str(project_id), run_id=str(run_id)
    )
    if isinstance(response, HTTPValidationError):
        raise ValueError(response.detail)
    if response is None:
        raise ValueError(f"Failed to get scorer jobs for project {project_id}, run {run_id}")

    _logger.debug(f"Scorer jobs: {response}")

    return [job for job in response if job.job_name == JobName.log_stream_scorer]


def scorer_jobs_status(project_id: str, run_id: str) -> None:
    """Gets the status of all scorer jobs for a given project and run.

    Parameters
    ----------
    project_id
        The unique identifier of the project.
    run_id
        The unique identifier of the run.
    """
    scorer_jobs = get_run_scorer_jobs(project_id, run_id)
    for job in scorer_jobs:
        scorer_name = None
        if "prompt_scorer_settings" in job.request_data:
            scorer_name = job.request_data["prompt_scorer_settings"]["scorer_name"]
        elif "scorer_config" in job.request_data:
            scorer_name = job.request_data["scorer_config"]["name"]

        if not scorer_name:
            _logger.debug(f"Scorer job {job.id} has no scorer name.")
            continue

        with contextlib.suppress(ValueError):
            scorer_name = Scorers(scorer_name).name

        _logger.debug(f"Scorer job {job.id} has scorer {scorer_name}.")

        if JobStatus.is_incomplete(job.status):
            _logger.info(f"{scorer_name.lstrip('_')}: Computing 🚧")
        elif JobStatus.is_failed(job.status):
            _logger.info(f"{scorer_name.lstrip('_')}: Failed ❌, error was: {job.error_message}")
        else:
            _logger.info(f"{scorer_name.lstrip('_')}: Done ✅")


def job_progress(job_id: str, project_id: str, run_id: str) -> UUID4:
    """Monitors the progress of a job and displays a progress bar.

    Parameters
    ----------
    job_id
        The unique identifier of the job to monitor.
    project_id
        The unique identifier of the project.
    run_id
        The unique identifier of the run.

    Returns
    -------
    The unique identifier of the completed job.
    """
    job_status = get_job(job_id)
    backoff = random.random()

    if JobStatus.is_incomplete(job_status.status):
        job_progress_bar = tqdm(total=job_status.steps_total, position=0, leave=True, desc=job_status.progress_message)
        while JobStatus.is_incomplete(job_status.status):
            sleep(backoff)
            job_status = get_job(job_id)
            job_progress_bar.set_description(job_status.progress_message)
            job_progress_bar.update(job_status.steps_completed - job_progress_bar.n)
            backoff = random.random()
        job_progress_bar.close()

    _logger.debug(f"Job {job_id} status: {job_status.status}.")
    if JobStatus.is_failed(job_status.status):
        raise ValueError(f"Job failed with error message {job_status.error_message}.") from None

    _logger.info("Initial job complete, executing scorers asynchronously. Current status:")
    scorer_jobs_status(project_id=project_id, run_id=run_id)
    return job_status.id



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.scorer_config import ScorerConfig
    from ..models.segment_filter import SegmentFilter


T = TypeVar("T", bound="MetricSettingsResponse")


@_attrs_define
class MetricSettingsResponse:
    """
    Attributes
    ----------
        scorers (list['ScorerConfig']):
        segment_filters (Union[None, Unset, list['SegmentFilter']]): List of segment filters to apply to the run.
    """

    scorers: list["ScorerConfig"]
    segment_filters: Union[None, Unset, list["SegmentFilter"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        scorers = []
        for scorers_item_data in self.scorers:
            scorers_item = scorers_item_data.to_dict()
            scorers.append(scorers_item)

        segment_filters: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.segment_filters, Unset):
            segment_filters = UNSET
        elif isinstance(self.segment_filters, list):
            segment_filters = []
            for segment_filters_type_0_item_data in self.segment_filters:
                segment_filters_type_0_item = segment_filters_type_0_item_data.to_dict()
                segment_filters.append(segment_filters_type_0_item)

        else:
            segment_filters = self.segment_filters

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"scorers": scorers})
        if segment_filters is not UNSET:
            field_dict["segment_filters"] = segment_filters

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.scorer_config import ScorerConfig
        from ..models.segment_filter import SegmentFilter

        d = dict(src_dict)
        scorers = []
        _scorers = d.pop("scorers")
        for scorers_item_data in _scorers:
            scorers_item = ScorerConfig.from_dict(scorers_item_data)

            scorers.append(scorers_item)

        def _parse_segment_filters(data: object) -> Union[None, Unset, list["SegmentFilter"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                segment_filters_type_0 = []
                _segment_filters_type_0 = data
                for segment_filters_type_0_item_data in _segment_filters_type_0:
                    segment_filters_type_0_item = SegmentFilter.from_dict(segment_filters_type_0_item_data)

                    segment_filters_type_0.append(segment_filters_type_0_item)

                return segment_filters_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["SegmentFilter"]], data)

        segment_filters = _parse_segment_filters(d.pop("segment_filters", UNSET))

        metric_settings_response = cls(scorers=scorers, segment_filters=segment_filters)

        metric_settings_response.additional_properties = d
        return metric_settings_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.log_records_boolean_filter_operator import LogRecordsBooleanFilterOperator
from ..types import UNSET, Unset

T = TypeVar("T", bound="LogRecordsBooleanFilter")


@_attrs_define
class LogRecordsBooleanFilter:
    """
    Attributes
    ----------
        column_id (str): ID of the column to filter.
        value (bool):
        operator (Union[Unset, LogRecordsBooleanFilterOperator]):  Default: LogRecordsBooleanFilterOperator.EQ.
        type_ (Union[Literal['boolean'], Unset]):  Default: 'boolean'.
    """

    column_id: str
    value: bool
    operator: Union[Unset, LogRecordsBooleanFilterOperator] = LogRecordsBooleanFilterOperator.EQ
    type_: Union[Literal["boolean"], Unset] = "boolean"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        column_id = self.column_id

        value = self.value

        operator: Union[Unset, str] = UNSET
        if not isinstance(self.operator, Unset):
            operator = self.operator.value

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"column_id": column_id, "value": value})
        if operator is not UNSET:
            field_dict["operator"] = operator
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        column_id = d.pop("column_id")

        value = d.pop("value")

        _operator = d.pop("operator", UNSET)
        operator: Union[Unset, LogRecordsBooleanFilterOperator]
        operator = UNSET if isinstance(_operator, Unset) else LogRecordsBooleanFilterOperator(_operator)

        type_ = cast(Union[Literal["boolean"], Unset], d.pop("type", UNSET))
        if type_ != "boolean" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'boolean', got '{type_}'")

        log_records_boolean_filter = cls(column_id=column_id, value=value, operator=operator, type_=type_)

        log_records_boolean_filter.additional_properties = d
        return log_records_boolean_filter

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any

from galileo.__future__.shared.base import StateManagementMixin, SyncState
from galileo.__future__.shared.exceptions import APIError, IntegrationNotConfiguredError, ValidationError
from galileo.config import GalileoPythonConfig
from galileo.resources.api.integrations import (
    create_or_update_integration_integrations_anthropic_put,
    create_or_update_integration_integrations_aws_bedrock_put,
    create_or_update_integration_integrations_azure_put,
    create_or_update_integration_integrations_openai_put,
    delete_integration_integrations_name_delete,
    get_integration_integrations_name_get,
)
from galileo.resources.api.llm_integrations import get_available_models_llm_integrations_llm_integration_models_get
from galileo.resources.models import (
    AnthropicIntegrationCreate,
    AzureIntegrationCreate,
    BaseAwsIntegrationCreate,
    HTTPValidationError,
    IntegrationName,
    OpenAIIntegrationCreate,
)
from galileo.resources.types import Unset
from galileo.utils.exceptions import APIException

logger = logging.getLogger(__name__)


class Provider(StateManagementMixin, ABC):
    """
    Base class for provider-specific integration objects.

    Providers represent configured integrations (e.g., OpenAI, Azure, Bedrock)
    that are stored remotely in the Galileo API. Providers are immutable proxies
    that only store minimal metadata (id, name, timestamps). No credentials are
    stored locally for security reasons.

    Providers should only be obtained through Integration class methods, not
    instantiated directly by users.

    Attributes
    ----------
        id (str | None): The unique integration identifier.
        name (str): The integration name/type (e.g., "openai", "azure").
        created_at (datetime | None): When the integration was created.
        updated_at (datetime | None): When the integration was last updated.
        created_by (str | None): The user who created the integration.
        is_selected (bool): Whether this integration is selected by the current user.
        permissions (list | None): Integration permissions for the current user.
    """

    id: str | None
    name: str
    created_at: datetime | None
    updated_at: datetime | None
    created_by: str | None
    is_selected: bool
    permissions: list[Any] | None

    def __init__(self) -> None:
        """Initialize a Provider instance."""
        super().__init__()
        self.id = None
        self.name = ""
        self.created_at = None
        self.updated_at = None
        self.created_by = None
        self.is_selected = False
        self.permissions = None
        self._set_state(SyncState.LOCAL_ONLY)

    def __str__(self) -> str:
        """String representation of the provider."""
        return f"{self.__class__.__name__}(name='{self.name}', id='{self.id}')"

    def __repr__(self) -> str:
        """Detailed string representation of the provider."""
        return f"{self.__class__.__name__}(name='{self.name}', id='{self.id}', is_selected={self.is_selected})"

    @abstractmethod
    def _get_integration_name(self) -> IntegrationName:
        """Get the IntegrationName enum for this provider."""
        raise NotImplementedError

    def refresh(self) -> None:
        """
        Refresh the provider state from the remote API.

        Raises
        ------
            APIError: If the API call fails or the integration is not found.
            ValidationError: If the provider has no ID.
        """
        if self.id is None:
            error = ValidationError("Cannot refresh provider without an ID")
            self._set_state(SyncState.FAILED_SYNC, error)
            raise error

        logger.info(f"{self.__class__.__name__}.refresh: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            integration_name = self._get_integration_name()

            # Get the specific integration data
            response = get_integration_integrations_name_get.sync(name=integration_name, client=config.api_client)

            if response is None or isinstance(response, HTTPValidationError):
                api_error = APIError(f"Provider with ID {self.id} not found")
                self._set_state(SyncState.FAILED_SYNC, api_error)
                raise api_error

            # Update attributes from response
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            logger.info(f"{self.__class__.__name__}.refresh: id='{self.id}' - completed")

        except APIException as e:
            error_msg = f"Failed to refresh provider: {e!s}"
            logger.error(f"{self.__class__.__name__}.refresh: id='{self.id}' - failed: {error_msg}")
            api_error = APIError(error_msg)
            self._set_state(SyncState.FAILED_SYNC, api_error)
            raise api_error from e

    def _update_from_api_response(self, response: Any) -> None:
        """Update instance attributes from API response."""
        self.id = str(response.id) if hasattr(response, "id") else self.id
        self.name = str(response.name) if hasattr(response, "name") else self.name
        self.created_at = response.created_at if hasattr(response, "created_at") else self.created_at
        self.updated_at = response.updated_at if hasattr(response, "updated_at") else self.updated_at
        self.created_by = response.created_by if hasattr(response, "created_by") else self.created_by

        # Handle Unset for optional fields
        if hasattr(response, "is_selected"):
            self.is_selected = response.is_selected if not isinstance(response.is_selected, Unset) else False
        if hasattr(response, "permissions"):
            self.permissions = response.permissions if not isinstance(response.permissions, Unset) else None

    def delete(self) -> None:
        """
        Delete this provider integration.

        This permanently removes the integration from the API. After deletion,
        the object state is set to DELETED.

        WARNING: This operation cannot be undone!

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been created yet.
        """
        if self.id is None:
            error = ValidationError("Cannot delete provider without an ID. Create the provider first.")
            self._set_state(SyncState.FAILED_SYNC, error)
            raise error

        logger.info(f"{self.__class__.__name__}.delete: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            integration_name = self._get_integration_name()

            result = delete_integration_integrations_name_delete.sync(name=integration_name, client=config.api_client)

            if isinstance(result, HTTPValidationError):
                raise APIError(f"Failed to delete provider: {result.detail}")

            self._set_state(SyncState.DELETED)
            logger.info(f"{self.__class__.__name__}.delete: id='{self.id}' - completed")

        except APIException as e:
            error_msg = f"Failed to delete provider: {e!s}"
            logger.error(f"{self.__class__.__name__}.delete: id='{self.id}' - failed: {error_msg}")
            api_error = APIError(error_msg)
            self._set_state(SyncState.FAILED_SYNC, api_error)
            raise api_error from e

    @property
    def models(self) -> list[Model]:
        """
        Get available models for this provider.

        Returns
        -------
            list[Model]: List of available models for this provider.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been synced with the API.
        """
        if not self.is_synced():
            error = ValidationError("Cannot get models for provider without syncing")
            raise error

        logger.info(f"{self.__class__.__name__}.models: Fetching models for '{self.name}'")

        try:
            config = GalileoPythonConfig.get()
            integration_name = self._get_integration_name()

            # Get models from API
            response = get_available_models_llm_integrations_llm_integration_models_get.sync(
                llm_integration=integration_name, client=config.api_client
            )

            if response is None or isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to get models for provider '{self.name}'")

            # Convert list of model name strings to Model objects
            models = [Model(name=model_name, alias=model_name, provider_name=self.name) for model_name in response]

            logger.info(f"{self.__class__.__name__}.models: Found {len(models)} model(s) for '{self.name}'")
            return models

        except APIException as e:
            error_msg = f"Failed to get models for provider: {e!s}"
            logger.error(f"{self.__class__.__name__}.models: failed: {error_msg}")
            raise APIError(error_msg) from e

    def get_model(self, *, name: str | None = None, alias: str | None = None) -> Model | None:
        """
        Get a specific model by name or alias.

        Args:
            name (str | None): The model name to search for.
            alias (str | None): The model alias to search for.

        Returns
        -------
            Model | None: The matching Model object, or None if not found.

        Raises
        ------
            ValidationError: If neither name nor alias is provided, or if both are provided.
            APIError: If the API call to fetch models fails.

        Examples
        --------
            # Get model by alias
            openai = Integration.openai
            model = openai.get_model(alias="gpt-4o-mini")

            # Get model by name
            model = openai.get_model(name="gpt-4o-mini")
        """
        if name is None and alias is None:
            raise ValidationError("Must specify either 'name' or 'alias'")
        if name is not None and alias is not None:
            raise ValidationError("Cannot specify both 'name' and 'alias'")

        logger.debug(
            f"{self.__class__.__name__}.get_model: "
            f"Searching for model with {'name' if name else 'alias'}="
            f"'{name if name else alias}' in provider '{self.name}'"
        )

        # Get all available models
        available_models = self.models

        # Search by name or alias
        search_value = name if name is not None else alias
        search_attr = "name" if name is not None else "alias"

        for model in available_models:
            if getattr(model, search_attr) == search_value:
                logger.debug(
                    f"{self.__class__.__name__}.get_model: Found model '{model.name}' with alias '{model.alias}'"
                )
                return model

        logger.debug(
            f"{self.__class__.__name__}.get_model: "
            f"No model found with {search_attr}='{search_value}' in provider '{self.name}'"
        )
        return None


class OpenAIProvider(Provider):
    """
    OpenAI integration provider.

    This is an immutable proxy to an OpenAI integration stored in the Galileo API.
    Credentials are never stored locally - they are only sent to the API during
    create/update operations.
    """

    # Temporary storage for credentials during creation (not persisted)
    _temp_token: str | None
    _temp_organization_id: str | None

    def __init__(self, *, token: str, organization_id: str | None = None) -> None:
        """
        Initialize an OpenAI provider for creation.

        Note: Credentials are only stored temporarily for the create() call.

        Args:
            token (str): OpenAI API token.
            organization_id (str | None): Optional organization ID.
        """
        super().__init__()
        self.name = "openai"
        # Store temporarily for create() call only
        self._temp_token = token
        self._temp_organization_id = organization_id

    def _get_integration_name(self) -> IntegrationName:
        return IntegrationName.OPENAI

    def create(self) -> OpenAIProvider:
        """
        Create or update this OpenAI integration in the API.

        Returns
        -------
            OpenAIProvider: Self, with updated state from API.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If credentials are not available.
        """
        if not hasattr(self, "_temp_token") or self._temp_token is None:
            raise ValidationError("Cannot create OpenAI provider without token")

        logger.info(f"OpenAIProvider.create: name='{self.name}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = OpenAIIntegrationCreate(token=self._temp_token, organization_id=self._temp_organization_id)

            response = create_or_update_integration_integrations_openai_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to create provider: {response.detail}")

            if response is None:
                raise APIError("Failed to create provider: no response")

            # Update attributes from response
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            # Clear temporary credentials
            self._temp_token = None
            self._temp_organization_id = None

            logger.info(f"OpenAIProvider.create: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"OpenAIProvider.create: failed: {e}")
            raise APIError(f"Failed to create OpenAI provider: {e}") from e

    def update(self, *, token: str, organization_id: str | None = None) -> OpenAIProvider:
        """
        Update this OpenAI integration credentials.

        Credentials are sent to the API but never stored locally.

        Args:
            token (str): New API token.
            organization_id (str | None): New organization ID.

        Returns
        -------
            OpenAIProvider: Self, with updated state.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been created yet.
        """
        if self.id is None:
            raise ValidationError("Cannot update provider without an ID. Create the provider first.")

        logger.info(f"OpenAIProvider.update: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = OpenAIIntegrationCreate(token=token, organization_id=organization_id)

            response = create_or_update_integration_integrations_openai_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to update provider: {response.detail}")

            if response is None:
                raise APIError("Failed to update provider: no response")

            # Update attributes from response (does NOT include credentials)
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            logger.info(f"OpenAIProvider.update: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"OpenAIProvider.update: failed: {e}")
            raise APIError(f"Failed to update OpenAI provider: {e}") from e


class AzureProvider(Provider):
    """
    Azure OpenAI integration provider.

    This is an immutable proxy to an Azure integration stored in the Galileo API.
    Credentials are never stored locally - they are only sent to the API during
    create/update operations.
    """

    # Temporary storage for credentials during creation (not persisted)
    _temp_token: str | None
    _temp_endpoint: str | None

    def __init__(self, *, token: str, endpoint: str) -> None:
        """
        Initialize an Azure provider for creation.

        Note: Credentials are only stored temporarily for the create() call.

        Args:
            token (str): Azure API key.
            endpoint (str): Azure OpenAI endpoint URL.
        """
        super().__init__()
        self.name = "azure"
        # Store temporarily for create() call only
        self._temp_token = token
        self._temp_endpoint = endpoint

    def _get_integration_name(self) -> IntegrationName:
        return IntegrationName.AZURE

    def create(self) -> AzureProvider:
        """
        Create or update this Azure integration in the API.

        Returns
        -------
            AzureProvider: Self, with updated state from API.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If credentials are not available.
        """
        if not hasattr(self, "_temp_token") or self._temp_token is None:
            raise ValidationError("Cannot create Azure provider without token")
        if not hasattr(self, "_temp_endpoint") or self._temp_endpoint is None:
            raise ValidationError("Cannot create Azure provider without endpoint")

        logger.info(f"AzureProvider.create: name='{self.name}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = AzureIntegrationCreate(token=self._temp_token, endpoint=self._temp_endpoint)

            response = create_or_update_integration_integrations_azure_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to create provider: {response.detail}")

            if response is None:
                raise APIError("Failed to create provider: no response")

            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            # Clear temporary credentials
            self._temp_token = None
            self._temp_endpoint = None

            logger.info(f"AzureProvider.create: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"AzureProvider.create: failed: {e}")
            raise APIError(f"Failed to create Azure provider: {e}") from e

    def update(self, *, token: str, endpoint: str) -> AzureProvider:
        """
        Update this Azure integration credentials.

        Credentials are sent to the API but never stored locally.

        Args:
            token (str): New API key.
            endpoint (str): New endpoint URL.

        Returns
        -------
            AzureProvider: Self, with updated state.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been created yet.
        """
        if self.id is None:
            raise ValidationError("Cannot update provider without an ID. Create the provider first.")

        logger.info(f"AzureProvider.update: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = AzureIntegrationCreate(token=token, endpoint=endpoint)

            response = create_or_update_integration_integrations_azure_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to update provider: {response.detail}")

            if response is None:
                raise APIError("Failed to update provider: no response")

            # Update attributes from response (does NOT include credentials)
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            logger.info(f"AzureProvider.update: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"AzureProvider.update: failed: {e}")
            raise APIError(f"Failed to update Azure provider: {e}") from e


class BedrockProvider(Provider):
    """
    AWS Bedrock integration provider.

    This is an immutable proxy to a Bedrock integration stored in the Galileo API.
    Credentials are never stored locally - they are only sent to the API during
    create/update operations.
    """

    # Temporary storage for credentials during creation (not persisted)
    _temp_credential_type: str | None
    _temp_region: str | None
    _temp_token_dict: dict[str, str] | None

    def __init__(
        self,
        *,
        credential_type: str = "key_secret",
        aws_access_key_id: str,
        aws_secret_access_key: str,
        region: str = "us-east-1",
    ) -> None:
        """
        Initialize a Bedrock provider for creation.

        Note: Credentials are only stored temporarily for the create() call.

        Args:
            credential_type (str): Type of credentials ("key_secret" or "assume_role").
            aws_access_key_id (str): AWS access key ID.
            aws_secret_access_key (str): AWS secret access key.
            region (str): AWS region.
        """
        super().__init__()
        self.name = "aws_bedrock"
        # Store temporarily for create() call only
        self._temp_credential_type = credential_type
        self._temp_region = region
        self._temp_token_dict = {"aws_access_key_id": aws_access_key_id, "aws_secret_access_key": aws_secret_access_key}

    def _get_integration_name(self) -> IntegrationName:
        return IntegrationName.AWS_BEDROCK

    def create(self) -> BedrockProvider:
        """
        Create or update this Bedrock integration in the API.

        Returns
        -------
            BedrockProvider: Self, with updated state from API.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If credentials are not available.
        """
        if not hasattr(self, "_temp_token_dict") or self._temp_token_dict is None:
            raise ValidationError("Cannot create Bedrock provider without credentials")

        logger.info(f"BedrockProvider.create: name='{self.name}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = BaseAwsIntegrationCreate(
                credential_type=self._temp_credential_type or "key_secret",
                token=self._temp_token_dict,
                region=self._temp_region or "us-east-1",
            )

            response = create_or_update_integration_integrations_aws_bedrock_put.sync(
                client=config.api_client, body=body
            )

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to create provider: {response.detail}")

            if response is None:
                raise APIError("Failed to create provider: no response")

            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            # Clear temporary credentials
            self._temp_credential_type = None
            self._temp_region = None
            self._temp_token_dict = None

            logger.info(f"BedrockProvider.create: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"BedrockProvider.create: failed: {e}")
            raise APIError(f"Failed to create Bedrock provider: {e}") from e

    def update(
        self,
        *,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        region: str = "us-east-1",
        credential_type: str = "key_secret",
    ) -> BedrockProvider:
        """
        Update this Bedrock integration credentials.

        Credentials are sent to the API but never stored locally.

        Args:
            aws_access_key_id (str): New AWS access key ID.
            aws_secret_access_key (str): New AWS secret access key.
            region (str): New AWS region.
            credential_type (str): Type of credentials.

        Returns
        -------
            BedrockProvider: Self, with updated state.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been created yet.
        """
        if self.id is None:
            raise ValidationError("Cannot update provider without an ID. Create the provider first.")

        logger.info(f"BedrockProvider.update: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            token_dict = {"aws_access_key_id": aws_access_key_id, "aws_secret_access_key": aws_secret_access_key}
            body = BaseAwsIntegrationCreate(credential_type=credential_type, token=token_dict, region=region)

            response = create_or_update_integration_integrations_aws_bedrock_put.sync(
                client=config.api_client, body=body
            )

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to update provider: {response.detail}")

            if response is None:
                raise APIError("Failed to update provider: no response")

            # Update attributes from response (does NOT include credentials)
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            logger.info(f"BedrockProvider.update: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"BedrockProvider.update: failed: {e}")
            raise APIError(f"Failed to update Bedrock provider: {e}") from e


class AnthropicProvider(Provider):
    """
    Anthropic (Claude) integration provider.

    This is an immutable proxy to an Anthropic integration stored in the Galileo API.
    Credentials are never stored locally - they are only sent to the API during
    create/update operations.
    """

    # Temporary storage for credentials during creation (not persisted)
    _temp_token: str | None

    def __init__(self, *, token: str) -> None:
        """
        Initialize an Anthropic provider for creation.

        Note: Credentials are only stored temporarily for the create() call.

        Args:
            token (str): Anthropic API token.
        """
        super().__init__()
        self.name = "anthropic"
        # Store temporarily for create() call only
        self._temp_token = token

    def _get_integration_name(self) -> IntegrationName:
        return IntegrationName.ANTHROPIC

    def create(self) -> AnthropicProvider:
        """
        Create or update this Anthropic integration in the API.

        Returns
        -------
            AnthropicProvider: Self, with updated state from API.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If credentials are not available.
        """
        if not hasattr(self, "_temp_token") or self._temp_token is None:
            raise ValidationError("Cannot create Anthropic provider without token")

        logger.info(f"AnthropicProvider.create: name='{self.name}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = AnthropicIntegrationCreate(token=self._temp_token)

            response = create_or_update_integration_integrations_anthropic_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to create provider: {response.detail}")

            if response is None:
                raise APIError("Failed to create provider: no response")

            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            # Clear temporary credentials
            self._temp_token = None

            logger.info(f"AnthropicProvider.create: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"AnthropicProvider.create: failed: {e}")
            raise APIError(f"Failed to create Anthropic provider: {e}") from e

    def update(self, *, token: str) -> AnthropicProvider:
        """
        Update this Anthropic integration credentials.

        Credentials are sent to the API but never stored locally.

        Args:
            token (str): New API token.

        Returns
        -------
            AnthropicProvider: Self, with updated state.

        Raises
        ------
            APIError: If the API call fails.
            ValidationError: If the provider has not been created yet.
        """
        if self.id is None:
            raise ValidationError("Cannot update provider without an ID. Create the provider first.")

        logger.info(f"AnthropicProvider.update: id='{self.id}' - started")

        try:
            config = GalileoPythonConfig.get()
            body = AnthropicIntegrationCreate(token=token)

            response = create_or_update_integration_integrations_anthropic_put.sync(client=config.api_client, body=body)

            if isinstance(response, HTTPValidationError):
                raise APIError(f"Failed to update provider: {response.detail}")

            if response is None:
                raise APIError("Failed to update provider: no response")

            # Update attributes from response (does NOT include credentials)
            self._update_from_api_response(response)
            self._set_state(SyncState.SYNCED)

            logger.info(f"AnthropicProvider.update: id='{self.id}' - completed")
            return self

        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"AnthropicProvider.update: failed: {e}")
            raise APIError(f"Failed to update Anthropic provider: {e}") from e


class GenericProvider(Provider):
    """
    Generic provider for integration types that don't have specialized implementations.

    This provider is used for integration types like 'mistral', 'nvidia', 'writer',
    'vertex_ai', 'aws_sagemaker', 'databricks', 'labelstudio', etc. that are available
    in the API but don't yet have dedicated Provider subclasses.

    GenericProvider is read-only and only supports basic operations:
    - Listing (returned from Integration.list())
    - Refreshing state
    - Deleting

    It does not support creation or updates through the SDK.
    """

    _integration_name: IntegrationName

    def _get_integration_name(self) -> IntegrationName:
        return self._integration_name


class UnconfiguredProvider:
    """
    Placeholder for integrations that are not configured.

    This class implements the Null Object Pattern to provide helpful error messages
    when users attempt to access an integration that doesn't exist. Instead of
    returning None (which leads to cryptic AttributeError messages), this class
    raises IntegrationNotConfiguredError with guidance on how to fix the issue.

    Examples
    --------
        # When Azure is not configured:
        >>> Integration.azure.models
        IntegrationNotConfiguredError: No 'azure' integration configured.
        Create one using Integration.create_azure() or configure it in the Galileo console.

        # Truthiness check still works:
        >>> if Integration.azure:
        ...     print("configured")
        ... else:
        ...     print("not configured")
        not configured
    """

    _integration_name: str

    def __init__(self, integration_name: str) -> None:
        # Use object.__setattr__ to bypass our custom __setattr__
        object.__setattr__(self, "_integration_name", integration_name)

    def __bool__(self) -> bool:
        """Allow truthiness checks: 'if Integration.azure:' returns False."""
        return False

    def __getattr__(self, name: str) -> None:
        """Raise helpful error when any attribute is accessed."""
        raise IntegrationNotConfiguredError(self._integration_name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Raise helpful error when any attribute is set."""
        raise IntegrationNotConfiguredError(self._integration_name)

    def __repr__(self) -> str:
        return f"UnconfiguredProvider('{self._integration_name}')"

    def __str__(self) -> str:
        return f"UnconfiguredProvider('{self._integration_name}')"


# Import Model here to avoid circular imports
from galileo.__future__.model import Model  # noqa: E402



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.output_type_enum import OutputTypeEnum
from ..types import UNSET, Unset

T = TypeVar("T", bound="GeneratedScorerConfiguration")


@_attrs_define
class GeneratedScorerConfiguration:
    """
    Attributes
    ----------
        model_alias (Union[Unset, str]):  Default: 'gpt-4.1-mini'.
        num_judges (Union[Unset, int]):  Default: 3.
        output_type (Union[Unset, OutputTypeEnum]): Enumeration of output types.
        scoreable_node_types (Union[Unset, list[str]]): Types of nodes that can be scored by this scorer.
        cot_enabled (Union[Unset, bool]): Whether chain of thought is enabled for this scorer. Default: False.
        ground_truth (Union[Unset, bool]): Whether ground truth is enabled for this scorer. Default: False.
    """

    model_alias: Union[Unset, str] = "gpt-4.1-mini"
    num_judges: Union[Unset, int] = 3
    output_type: Union[Unset, OutputTypeEnum] = UNSET
    scoreable_node_types: Union[Unset, list[str]] = UNSET
    cot_enabled: Union[Unset, bool] = False
    ground_truth: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        model_alias = self.model_alias

        num_judges = self.num_judges

        output_type: Union[Unset, str] = UNSET
        if not isinstance(self.output_type, Unset):
            output_type = self.output_type.value

        scoreable_node_types: Union[Unset, list[str]] = UNSET
        if not isinstance(self.scoreable_node_types, Unset):
            scoreable_node_types = self.scoreable_node_types

        cot_enabled = self.cot_enabled

        ground_truth = self.ground_truth

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if model_alias is not UNSET:
            field_dict["model_alias"] = model_alias
        if num_judges is not UNSET:
            field_dict["num_judges"] = num_judges
        if output_type is not UNSET:
            field_dict["output_type"] = output_type
        if scoreable_node_types is not UNSET:
            field_dict["scoreable_node_types"] = scoreable_node_types
        if cot_enabled is not UNSET:
            field_dict["cot_enabled"] = cot_enabled
        if ground_truth is not UNSET:
            field_dict["ground_truth"] = ground_truth

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        model_alias = d.pop("model_alias", UNSET)

        num_judges = d.pop("num_judges", UNSET)

        _output_type = d.pop("output_type", UNSET)
        output_type: Union[Unset, OutputTypeEnum]
        output_type = UNSET if isinstance(_output_type, Unset) else OutputTypeEnum(_output_type)

        scoreable_node_types = cast(list[str], d.pop("scoreable_node_types", UNSET))

        cot_enabled = d.pop("cot_enabled", UNSET)

        ground_truth = d.pop("ground_truth", UNSET)

        generated_scorer_configuration = cls(
            model_alias=model_alias,
            num_judges=num_judges,
            output_type=output_type,
            scoreable_node_types=scoreable_node_types,
            cot_enabled=cot_enabled,
            ground_truth=ground_truth,
        )

        generated_scorer_configuration.additional_properties = d
        return generated_scorer_configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Callable


class classproperty:
    """Decorator for class-level properties."""

    def __init__(self, func: Callable[[Any], Any]) -> None:
        self.func = func
        self.__doc__ = func.__doc__

    def __get__(self, obj: Any, owner: Any) -> Any:
        return self.func(owner)

    def __set__(self, obj: Any, value: Any) -> None:
        raise AttributeError("can't set attribute")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.experiment_phase_status import ExperimentPhaseStatus


T = TypeVar("T", bound="ExperimentStatus")


@_attrs_define
class ExperimentStatus:
    """
    Attributes
    ----------
        log_generation (Union[Unset, ExperimentPhaseStatus]):
    """

    log_generation: Union[Unset, "ExperimentPhaseStatus"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        log_generation: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.log_generation, Unset):
            log_generation = self.log_generation.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if log_generation is not UNSET:
            field_dict["log_generation"] = log_generation

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.experiment_phase_status import ExperimentPhaseStatus

        d = dict(src_dict)
        _log_generation = d.pop("log_generation", UNSET)
        log_generation: Union[Unset, ExperimentPhaseStatus]
        if isinstance(_log_generation, Unset):
            log_generation = UNSET
        else:
            log_generation = ExperimentPhaseStatus.from_dict(_log_generation)

        experiment_status = cls(log_generation=log_generation)

        experiment_status.additional_properties = d
        return experiment_status

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from io import BytesIO
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from .. import types
from ..types import UNSET, File, Unset

T = TypeVar("T", bound="BodyValidateCodeScorerScorersCodeValidatePost")


@_attrs_define
class BodyValidateCodeScorerScorersCodeValidatePost:
    """
    Attributes
    ----------
        file (File):
        required_scorers (Union[None, Unset, list[str], str]):
        scoreable_node_types (Union[None, Unset, list[str], str]):
    """

    file: File
    required_scorers: Union[None, Unset, list[str], str] = UNSET
    scoreable_node_types: Union[None, Unset, list[str], str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        file = self.file.to_tuple()

        required_scorers: Union[None, Unset, list[str], str]
        if isinstance(self.required_scorers, Unset):
            required_scorers = UNSET
        elif isinstance(self.required_scorers, list):
            required_scorers = self.required_scorers

        else:
            required_scorers = self.required_scorers

        scoreable_node_types: Union[None, Unset, list[str], str]
        if isinstance(self.scoreable_node_types, Unset):
            scoreable_node_types = UNSET
        elif isinstance(self.scoreable_node_types, list):
            scoreable_node_types = self.scoreable_node_types

        else:
            scoreable_node_types = self.scoreable_node_types

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"file": file})
        if required_scorers is not UNSET:
            field_dict["required_scorers"] = required_scorers
        if scoreable_node_types is not UNSET:
            field_dict["scoreable_node_types"] = scoreable_node_types

        return field_dict

    def to_multipart(self) -> types.RequestFiles:
        files: types.RequestFiles = []

        files.append(("file", self.file.to_tuple()))

        if not isinstance(self.required_scorers, Unset):
            if isinstance(self.required_scorers, str):
                files.append(("required_scorers", (None, str(self.required_scorers).encode(), "text/plain")))
            elif isinstance(self.required_scorers, list):
                for required_scorers_type_1_item_element in self.required_scorers:
                    files.append(
                        ("required_scorers", (None, str(required_scorers_type_1_item_element).encode(), "text/plain"))
                    )
            else:
                files.append(("required_scorers", (None, str(self.required_scorers).encode(), "text/plain")))

        if not isinstance(self.scoreable_node_types, Unset):
            if isinstance(self.scoreable_node_types, str):
                files.append(("scoreable_node_types", (None, str(self.scoreable_node_types).encode(), "text/plain")))
            elif isinstance(self.scoreable_node_types, list):
                for scoreable_node_types_type_1_item_element in self.scoreable_node_types:
                    files.append(
                        (
                            "scoreable_node_types",
                            (None, str(scoreable_node_types_type_1_item_element).encode(), "text/plain"),
                        )
                    )
            else:
                files.append(("scoreable_node_types", (None, str(self.scoreable_node_types).encode(), "text/plain")))

        for prop_name, prop in self.additional_properties.items():
            files.append((prop_name, (None, str(prop).encode(), "text/plain")))

        return files

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        file = File(payload=BytesIO(d.pop("file")))

        def _parse_required_scorers(data: object) -> Union[None, Unset, list[str], str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                return cast(list[str], data)

            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str], str], data)

        required_scorers = _parse_required_scorers(d.pop("required_scorers", UNSET))

        def _parse_scoreable_node_types(data: object) -> Union[None, Unset, list[str], str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                return cast(list[str], data)

            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str], str], data)

        scoreable_node_types = _parse_scoreable_node_types(d.pop("scoreable_node_types", UNSET))

        body_validate_code_scorer_scorers_code_validate_post = cls(
            file=file, required_scorers=required_scorers, scoreable_node_types=scoreable_node_types
        )

        body_validate_code_scorer_scorers_code_validate_post.additional_properties = d
        return body_validate_code_scorer_scorers_code_validate_post

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.few_shot_example import FewShotExample
    from ..models.groundedness_template_response_schema_type_0 import GroundednessTemplateResponseSchemaType0


T = TypeVar("T", bound="GroundednessTemplate")


@_attrs_define
class GroundednessTemplate:
    r"""Template for the groundedness metric,
    containing all the info necessary to send the groundedness prompt.

    Attributes
    ----------
            metric_system_prompt (Union[Unset, str]):  Default: 'The user will provide you with a prompt that was sent to an
                automatic question-answering system, and that system\'s response. Both will be provided as JSON strings.\n\nThe
                prompt will contain one or more documents intended as context which the question-answering system was given as
                reference material.\n\nYour task is to determine whether the answer was supported by the documents.\n\nThink
                step by step, and explain your reasoning carefully.\nState your observations first, before drawing any
                conclusions.\n\nRespond in the following JSON format:\n\n```\n{\n    \\"explanation\\": string,\n
                \\"was_supported\\": boolean\n}\n```\n\n\\"explanation\\": Your step-by-step reasoning process. List out the
                claims made in the response, and for each claim, provide a detailed explanation of why that claim is or is not
                supported by the documents.\n\n\\"was_supported\\": `true` if the response was supported by the documents,
                `false` otherwise.\n\nYou must respond with valid JSON.'.
            metric_description (Union[Unset, str]):  Default: 'I have a RAG (retrieval-augmented generation) system that
                generates text based on one or more documents that I always include in my prompts. I want a metric that checks
                whether the generated text was supported by information in the documents. The metric should exhaustively check
                each claim in the response against the documents, one by one, listing them out explicitly.'.
            value_field_name (Union[Unset, str]):  Default: 'was_supported'.
            explanation_field_name (Union[Unset, str]): Field name to look for in the chainpoll response, for the
                explanation. Default: 'explanation'.
            template (Union[Unset, str]):  Default: 'Prompt JSON:\n\n```\n{query_json}\n```\n\nResponse
                JSON:\n\n```\n{response_json}\n```'.
            metric_few_shot_examples (Union[Unset, list['FewShotExample']]):
            response_schema (Union['GroundednessTemplateResponseSchemaType0', None, Unset]): Response schema for the output
    """

    metric_system_prompt: Union[Unset, str] = (
        'The user will provide you with a prompt that was sent to an automatic question-answering system, and that system\'s response. Both will be provided as JSON strings.\n\nThe prompt will contain one or more documents intended as context which the question-answering system was given as reference material.\n\nYour task is to determine whether the answer was supported by the documents.\n\nThink step by step, and explain your reasoning carefully.\nState your observations first, before drawing any conclusions.\n\nRespond in the following JSON format:\n\n```\n{\n    \\"explanation\\": string,\n    \\"was_supported\\": boolean\n}\n```\n\n\\"explanation\\": Your step-by-step reasoning process. List out the claims made in the response, and for each claim, provide a detailed explanation of why that claim is or is not supported by the documents.\n\n\\"was_supported\\": `true` if the response was supported by the documents, `false` otherwise.\n\nYou must respond with valid JSON.'
    )
    metric_description: Union[Unset, str] = (
        "I have a RAG (retrieval-augmented generation) system that generates text based on one or more documents that I always include in my prompts. I want a metric that checks whether the generated text was supported by information in the documents. The metric should exhaustively check each claim in the response against the documents, one by one, listing them out explicitly."
    )
    value_field_name: Union[Unset, str] = "was_supported"
    explanation_field_name: Union[Unset, str] = "explanation"
    template: Union[Unset, str] = (
        "Prompt JSON:\n\n```\n{query_json}\n```\n\nResponse JSON:\n\n```\n{response_json}\n```"
    )
    metric_few_shot_examples: Union[Unset, list["FewShotExample"]] = UNSET
    response_schema: Union["GroundednessTemplateResponseSchemaType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.groundedness_template_response_schema_type_0 import GroundednessTemplateResponseSchemaType0

        metric_system_prompt = self.metric_system_prompt

        metric_description = self.metric_description

        value_field_name = self.value_field_name

        explanation_field_name = self.explanation_field_name

        template = self.template

        metric_few_shot_examples: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.metric_few_shot_examples, Unset):
            metric_few_shot_examples = []
            for metric_few_shot_examples_item_data in self.metric_few_shot_examples:
                metric_few_shot_examples_item = metric_few_shot_examples_item_data.to_dict()
                metric_few_shot_examples.append(metric_few_shot_examples_item)

        response_schema: Union[None, Unset, dict[str, Any]]
        if isinstance(self.response_schema, Unset):
            response_schema = UNSET
        elif isinstance(self.response_schema, GroundednessTemplateResponseSchemaType0):
            response_schema = self.response_schema.to_dict()
        else:
            response_schema = self.response_schema

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if metric_system_prompt is not UNSET:
            field_dict["metric_system_prompt"] = metric_system_prompt
        if metric_description is not UNSET:
            field_dict["metric_description"] = metric_description
        if value_field_name is not UNSET:
            field_dict["value_field_name"] = value_field_name
        if explanation_field_name is not UNSET:
            field_dict["explanation_field_name"] = explanation_field_name
        if template is not UNSET:
            field_dict["template"] = template
        if metric_few_shot_examples is not UNSET:
            field_dict["metric_few_shot_examples"] = metric_few_shot_examples
        if response_schema is not UNSET:
            field_dict["response_schema"] = response_schema

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.few_shot_example import FewShotExample
        from ..models.groundedness_template_response_schema_type_0 import GroundednessTemplateResponseSchemaType0

        d = dict(src_dict)
        metric_system_prompt = d.pop("metric_system_prompt", UNSET)

        metric_description = d.pop("metric_description", UNSET)

        value_field_name = d.pop("value_field_name", UNSET)

        explanation_field_name = d.pop("explanation_field_name", UNSET)

        template = d.pop("template", UNSET)

        metric_few_shot_examples = []
        _metric_few_shot_examples = d.pop("metric_few_shot_examples", UNSET)
        for metric_few_shot_examples_item_data in _metric_few_shot_examples or []:
            metric_few_shot_examples_item = FewShotExample.from_dict(metric_few_shot_examples_item_data)

            metric_few_shot_examples.append(metric_few_shot_examples_item)

        def _parse_response_schema(data: object) -> Union["GroundednessTemplateResponseSchemaType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return GroundednessTemplateResponseSchemaType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["GroundednessTemplateResponseSchemaType0", None, Unset], data)

        response_schema = _parse_response_schema(d.pop("response_schema", UNSET))

        groundedness_template = cls(
            metric_system_prompt=metric_system_prompt,
            metric_description=metric_description,
            value_field_name=value_field_name,
            explanation_field_name=explanation_field_name,
            template=template,
            metric_few_shot_examples=metric_few_shot_examples,
            response_schema=response_schema,
        )

        groundedness_template.additional_properties = d
        return groundedness_template

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum
from typing import Literal, Optional, Union

from pydantic import UUID4, BaseModel, Field

from galileo.resources.models import Document
from galileo_core.schemas.logging.span import Span
from galileo_core.schemas.logging.step import StepAllowedInputType, StepAllowedOutputType
from galileo_core.schemas.logging.trace import Trace

SPAN_TYPE = Literal["llm", "retriever", "tool", "workflow", "agent"]


class LoggingMethod(str, Enum):
    playground = "playground"
    python_client = "python_client"
    typescript_client = "typescript_client"
    api_direct = "api_direct"


class BaseLogStreamOrExperimentModel(BaseModel):
    log_stream_id: Optional[UUID4] = Field(default=None, description="Log stream id associated with the traces.")
    experiment_id: Optional[UUID4] = Field(default=None, description="Experiment id associated with the traces.")


class LogRecordsIngestRequest(BaseLogStreamOrExperimentModel):
    logging_method: LoggingMethod = Field(default=LoggingMethod.api_direct)
    client_version: Union[str, None] = Field(default=None)
    reliable: bool = Field(
        default=False,
        description="Whether or not to use reliable logging.  If set to False, the method will respond immediately before verifying that the traces have been successfully ingested, and no error message will be returned if ingestion fails.  If set to True, the method will wait for the traces to be successfully ingested or return an error message if there is an ingestion failure.",
    )


class TracesIngestRequest(LogRecordsIngestRequest):
    traces: list[Trace] = Field(..., description="List of traces to log.", min_length=1)
    session_id: Optional[UUID4] = Field(default=None, description="Session id associated with the traces.")
    is_complete: Optional[bool] = Field(default=True, description="Is complete.")


class SpansIngestRequest(LogRecordsIngestRequest):
    spans: list[Span] = Field(..., description="List of spans to log.", min_length=1)
    trace_id: UUID4 = Field(description="Trace id associated with the spans.")
    parent_id: UUID4 = Field(description="Parent trace or span id.")


class LogRecordsIngestResponse(BaseLogStreamOrExperimentModel):
    project_id: UUID4 = Field(description="Project id associated with the traces.")
    project_name: str = Field(description="Project name associated with the traces.")
    session_id: UUID4 = Field(description="Session id associated with the traces.")
    records_count: int = Field(description="Total number of records ingested")


class TracesIngestResponse(LogRecordsIngestResponse):
    traces_count: int = Field(description="total number of traces ingested")


class LogTraceUpdateResponse(LogRecordsIngestResponse):
    trace_id: UUID4 = Field(description="Trace id associated with the updated trace.")


class LogSpansIngestResponse(LogRecordsIngestResponse):
    trace_id: UUID4 = Field(description="Trace id associated with the spans.")
    parent_id: UUID4 = Field(description="Parent trace or span id.")


class LogSpanUpdateResponse(LogRecordsIngestResponse):
    span_id: UUID4 = Field(description="Span id associated with the updated span.")


class TraceUpdateRequest(LogRecordsIngestRequest):
    trace_id: UUID4 = Field(..., description="Trace id to update.")
    input: Union[str, None] = Field(
        default=None, description="Input of the trace. Overwrites previous value if present."
    )
    output: Union[str, None] = Field(
        default=None, description="Output of the trace. Overwrites previous value if present."
    )
    status_code: Union[int, None] = Field(
        default=None, description="Status code of the trace. Overwrites previous value if present."
    )
    tags: Union[list[str], None] = Field(default=None, description="Tags to add to the trace.")
    is_complete: Union[bool, None] = Field(
        default=False, description="Whether or not the records in this request are complete."
    )
    duration_ns: Union[int, None] = Field(
        default=None, description="Duration in nanoseconds. Overwrites previous value if present."
    )


class SpanUpdateRequest(LogRecordsIngestRequest):
    span_id: UUID4 = Field(..., description="Span id to update.")
    input: Union[StepAllowedInputType, None] = Field(
        default=None, description="Input of the span. Overwrites previous value if present."
    )
    output: Union[StepAllowedOutputType, None] = Field(
        default=None, description="Output of the span. Overwrites previous value if present."
    )
    tags: Union[list[str], None] = Field(default=None, description="Tags to add to the span.")
    status_code: Union[int, None] = Field(
        default=None, description="Status code of the span. Overwrites previous value if present."
    )
    duration_ns: Union[int, None] = Field(
        default=None, description="Duration in nanoseconds. Overwrites previous value if present."
    )


class SessionCreateRequest(BaseLogStreamOrExperimentModel):
    name: Optional[str] = Field(default=None, description="Name of the session.")
    previous_session_id: Optional[UUID4] = Field(default=None, description="Previous session id.")
    external_id: Optional[str] = Field(default=None, description="External id of the session.")


class SessionCreateResponse(BaseLogStreamOrExperimentModel):
    id: UUID4 = Field(description="Id of the session.")
    name: Optional[str] = Field(default=None, description="Name of the session.")
    previous_session_id: Optional[UUID4] = Field(default=None, description="Previous session id.")
    external_id: Optional[str] = Field(default=None, description="External id of the session.")
    project_id: UUID4 = Field(description="Project id associated with the session.")
    project_name: str = Field(description="Project name associated with the session.")
    log_stream_id: Optional[UUID4] = Field(default=None, description="Log stream id associated with the session.")


class LogRecordsSearchFilterOperator(str, Enum):
    eq = "eq"
    ne = "ne"
    contains = "contains"
    one_of = "one_of"
    not_in = "not_in"
    gt = "gt"
    gte = "gte"
    lt = "lt"
    lte = "lte"
    between = "between"


class LogRecordsSearchFilterType(str, Enum):
    id = "id"
    date = "date"
    number = "number"
    boolean = "boolean"
    text = "text"


class LogRecordsSearchFilter(BaseModel):
    type: LogRecordsSearchFilterType = Field(description="Type of the log records filter.")
    name: str = Field(description="ID of the column to filter.", alias="column_id")
    value: str = Field(description="Value to filter by.")
    operator: LogRecordsSearchFilterOperator = Field(description="Operator to apply to the filter.")


class LogRecordsSearchRequest(BaseLogStreamOrExperimentModel):
    filters: Optional[list[LogRecordsSearchFilter]] = Field(default=None, description="Filters to apply to the search.")


RetrieverSpanAllowedOutputType = Union[
    str, list[str], dict[str, str], list[dict[str, str]], Document, list[Document], None
]



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from galileo.utils.headers_data import get_sdk_header
from galileo_core.constants.request_method import RequestMethod
from galileo_core.helpers.api_client import ApiClient

from ... import errors
from ...models.http_validation_error import HTTPValidationError
from ...models.run_tag_db import RunTagDB
from ...types import Response


def _get_kwargs(project_id: str, experiment_id: str, tag_id: str) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": RequestMethod.GET,
        "return_raw_response": True,
        "path": f"/projects/{project_id}/experiments/{experiment_id}/tags/{tag_id}",
    }

    headers["X-Galileo-SDK"] = get_sdk_header()

    _kwargs["content_headers"] = headers
    return _kwargs


def _parse_response(*, client: ApiClient, response: httpx.Response) -> Optional[Union[HTTPValidationError, RunTagDB]]:
    if response.status_code == 200:
        return RunTagDB.from_dict(response.json())

    if response.status_code == 422:
        return HTTPValidationError.from_dict(response.json())

    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    return None


def _build_response(*, client: ApiClient, response: httpx.Response) -> Response[Union[HTTPValidationError, RunTagDB]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str, experiment_id: str, tag_id: str, *, client: ApiClient
) -> Response[Union[HTTPValidationError, RunTagDB]]:
    """Get Experiment Tag.

     Gets a tag for a given project_id/experiment_id.

    Args:
        project_id (str):
        experiment_id (str):
        tag_id (str):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, RunTagDB]]
    """
    kwargs = _get_kwargs(project_id=project_id, experiment_id=experiment_id, tag_id=tag_id)

    response = client.request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    project_id: str, experiment_id: str, tag_id: str, *, client: ApiClient
) -> Optional[Union[HTTPValidationError, RunTagDB]]:
    """Get Experiment Tag.

     Gets a tag for a given project_id/experiment_id.

    Args:
        project_id (str):
        experiment_id (str):
        tag_id (str):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, RunTagDB]
    """
    return sync_detailed(project_id=project_id, experiment_id=experiment_id, tag_id=tag_id, client=client).parsed


async def asyncio_detailed(
    project_id: str, experiment_id: str, tag_id: str, *, client: ApiClient
) -> Response[Union[HTTPValidationError, RunTagDB]]:
    """Get Experiment Tag.

     Gets a tag for a given project_id/experiment_id.

    Args:
        project_id (str):
        experiment_id (str):
        tag_id (str):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, RunTagDB]]
    """
    kwargs = _get_kwargs(project_id=project_id, experiment_id=experiment_id, tag_id=tag_id)

    response = await client.arequest(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str, experiment_id: str, tag_id: str, *, client: ApiClient
) -> Optional[Union[HTTPValidationError, RunTagDB]]:
    """Get Experiment Tag.

     Gets a tag for a given project_id/experiment_id.

    Args:
        project_id (str):
        experiment_id (str):
        tag_id (str):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, RunTagDB]
    """
    return (
        await asyncio_detailed(project_id=project_id, experiment_id=experiment_id, tag_id=tag_id, client=client)
    ).parsed



from __future__ import annotations

import builtins
import json
import logging
import os
import time
from abc import ABC
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from galileo.__future__.model import Model

from galileo.__future__.configuration import Configuration
from galileo.__future__.shared.base import StateManagementMixin, SyncState
from galileo.__future__.shared.exceptions import ValidationError
from galileo.config import GalileoPythonConfig
from galileo.metrics import Metrics
from galileo.resources.api.data import (
    create_code_scorer_version_scorers_scorer_id_version_code_post,
    create_scorers_post,
    get_validate_code_scorer_task_result_scorers_code_validate_task_id_get,
    validate_code_scorer_scorers_code_validate_post,
)
from galileo.resources.models import (
    BodyCreateCodeScorerVersionScorersScorerIdVersionCodePost,
    BodyValidateCodeScorerScorersCodeValidatePost,
    CreateScorerRequest,
    OutputTypeEnum,
    ScorerTypes,
    TaskResultStatus,
)
from galileo.resources.models.invalid_result import InvalidResult
from galileo.resources.types import File, Unset
from galileo.schema.metrics import GalileoMetrics, LocalMetricConfig
from galileo.schema.metrics import Metric as LegacyMetric
from galileo.scorers import Scorers
from galileo_core.schemas.logging.span import Span
from galileo_core.schemas.logging.step import StepType
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.metric import MetricValueType

logger = logging.getLogger(__name__)

# Code validation polling parameters are configurable via Configuration:
#   - Configuration.code_validation_timeout (env: GALILEO_CODE_VALIDATION_TIMEOUT) - default: 60.0s
#   - Configuration.code_validation_initial_delay (env: GALILEO_CODE_VALIDATION_INITIAL_DELAY) - default: 5.0s
#   - Configuration.code_validation_max_delay (env: GALILEO_CODE_VALIDATION_MAX_DELAY) - default: 30.0s
#   - Configuration.code_validation_backoff_multiplier (env: GALILEO_CODE_VALIDATION_BACKOFF_MULTIPLIER) - default: 1.5


class BuiltInMetrics:
    """
    Provides convenient access to built-in Galileo metrics (formerly "scorers").

    Examples
    --------
        from galileo.__future__ import Metric

        # Access built-in metrics
        Metric.metrics.correctness
        Metric.metrics.completeness
        Metric.metrics.toxicity
    """

    def __getattr__(self, name: str) -> GalileoMetrics:
        """Allow attribute-style access to built-in metrics."""
        # Try to find the metric by name (enum names match UI-visible names)
        for scorer in GalileoMetrics:
            if scorer.name == name:
                return scorer
        raise AttributeError(f"Built-in metric '{name}' not found. Available: {[s.name for s in GalileoMetrics]}")

    def __dir__(self) -> list[str]:
        """Return list of available metric names for autocomplete."""
        return [scorer.name for scorer in GalileoMetrics]


# Backwards-compatible alias
BuiltInScorers = BuiltInMetrics


class Metric(StateManagementMixin, ABC):
    """
    Base class for all Galileo metrics.

    This is an abstract base class that defines common attributes and methods
    for all metric types. Use one of the concrete metric classes instead:

    - **GalileoMetric**: Built-in Galileo scorers (access via Metric.scorers)
    - **LlmMetric**: Custom LLM-based metrics with prompt templates
    - **LocalMetric**: Local function-based metrics
    - **CodeMetric**: Code-based metrics (future support)

    Common Attributes
    -----------------
        id (str | None): The unique metric identifier (UUID).
        name (str): The metric name.
        scorer_type (ScorerTypes | None): The type of scorer.
        description (str): Description of the metric.
        tags (list[str]): Tags associated with the metric.
        created_at (datetime | None): When the metric was created.
        updated_at (datetime | None): When the metric was last updated.
        version (int | None): Metric version number.

    Class Attributes
    ----------------
        metrics (BuiltInMetrics): Access built-in Galileo metrics.

    Examples
    --------
        # 1. Use built-in Galileo scorers
        from galileo.__future__ import Metric, GalileoMetric, LlmMetric, LocalMetric, LogStream

        log_stream = LogStream.get(name="my-stream", project_name="my-project")
        log_stream.set_metrics([
            Metric.metrics.correctness,
            Metric.metrics.completeness,
        ])

        # 2. Create custom LLM metric
        llm_metric = LlmMetric(
            name="response_quality",
            prompt="Rate the quality...",
            model="gpt-4o-mini",
            judges=3,
        ).create()

        # 3. Create local function-based metric
        def my_scorer(trace_or_span):
            return 0.5

        local_metric = LocalMetric(
            name="response_length",
            scorer_fn=my_scorer,
        )
    """

    # Class attribute for built-in metrics (preferred name)
    metrics = BuiltInMetrics()

    # Backwards-compatible property for legacy name
    scorers = metrics

    # Type annotations for common instance attributes
    id: str | None
    name: str
    scorer_type: ScorerTypes | None
    description: str
    tags: list[str]
    created_at: datetime | None
    updated_at: datetime | None
    version: int | None

    # Scorer defaults - available for LLM and built-in Galileo metrics
    # These are returned by the API in the ScorerDefaults object
    model: str | None
    judges: int | None
    cot_enabled: bool | None

    def __init__(
        self, name: str, *, description: str = "", tags: list[str] | None = None, version: int | None = None
    ) -> None:
        """
        Initialize a base Metric instance with common attributes.

        Args:
            name: The name of the metric.
            description: Description of the metric.
            tags: Tags associated with the metric.
            version: Specific version to reference (for existing metrics).
        """
        super().__init__()
        self.name = name
        self.description = description
        self.tags = tags if tags is not None else []
        self.version = version
        self.id = None
        self.created_at = None
        self.updated_at = None
        self.scorer_type = None

        # Initialize scorer defaults (populated from API for LLM and Galileo metrics)
        self.model = None
        self.judges = None
        self.cot_enabled = None

        self._set_state(SyncState.LOCAL_ONLY)

    @classmethod
    def _create_metric_from_type(cls, scorer_type: ScorerTypes) -> Metric:
        """
        Create the appropriate Metric subclass instance based on scorer_type.

        This is a factory method that centralizes the logic for instantiating
        the correct metric subclass based on the scorer type returned from the API.

        Args:
            scorer_type: The scorer type from the API response.

        Returns
        -------
            Metric: An uninitialized instance of the appropriate subclass
                   (LlmMetric, CodeMetric, or GalileoMetric).

        Examples
        --------
            instance = Metric._create_metric_from_type(ScorerTypes.LLM)
            # Returns: LlmMetric instance
        """
        if scorer_type == ScorerTypes.LLM:
            return LlmMetric.__new__(LlmMetric)
        if scorer_type == ScorerTypes.CODE:
            return CodeMetric.__new__(CodeMetric)
        # Default to GalileoMetric for built-in scorers (LUNA, PRESET, etc.)
        return GalileoMetric.__new__(GalileoMetric)

    @classmethod
    def get(cls, *, id: str | None = None, name: str | None = None) -> Metric | None:
        """
        Get an existing metric by ID or name.

        Returns the appropriate subclass instance based on scorer_type.

        Args:
            id: The metric ID (UUID).
            name: The metric name.

        Returns
        -------
            Optional[Metric]: The metric if found (GalileoMetric, LlmMetric, or CodeMetric), None otherwise.

        Raises
        ------
            ValidationError: If neither or both id and name are provided.

        Examples
        --------
            # Get by name - returns appropriate subclass
            metric = Metric.get(name="factuality-checker")

            # Get by ID
            metric = Metric.get(id="abc-123-def")
        """
        if id is not None and name is not None:
            raise ValidationError("Cannot specify both id and name")
        if id is None and name is None:
            raise ValidationError("Must specify either id or name")

        scorers_service = Scorers()

        if name is not None:
            scorers = scorers_service.list(name=name)
            if not scorers:
                return None
            retrieved_scorer = next((s for s in scorers if s.name == name), None)
            if retrieved_scorer is None:
                return None
        else:
            assert id is not None
            scorers = scorers_service.list()
            retrieved_scorer = next((s for s in scorers if s.id == id), None)
            if retrieved_scorer is None:
                return None

        # Create appropriate subclass instance based on scorer_type
        instance = cls._create_metric_from_type(retrieved_scorer.scorer_type)
        StateManagementMixin.__init__(instance)
        instance._populate_from_scorer_response(retrieved_scorer)
        instance._set_state(SyncState.SYNCED)
        return instance

    @classmethod
    def list(
        cls, *, name_filter: str | None = None, scorer_types: list[ScorerTypes] | None = None
    ) -> builtins.list[Metric]:
        """
        List metrics with optional filtering.

        Returns appropriate subclass instances based on scorer_type.

        Args:
            name_filter: Filter metrics by exact name match.
            scorer_types: Filter by scorer types.

        Returns
        -------
            list[Metric]: List of metrics matching the criteria (with appropriate subclass types).

        Examples
        --------
            # List all metrics
            metrics = Metric.list()

            # List LLM metrics only
            metrics = Metric.list(scorer_types=[ScorerTypes.LLM])

            # List by name
            metrics = Metric.list(name_filter="factuality")
        """
        logger.debug(f"Metric.list: name_filter='{name_filter}' types={scorer_types} - started")
        scorers_service = Scorers()
        retrieved_scorers = scorers_service.list(name=name_filter, types=scorer_types)
        logger.debug(f"Metric.list: found {len(retrieved_scorers)} metrics - completed")

        result: builtins.list[Metric] = []
        for retrieved_scorer in retrieved_scorers:
            # Create appropriate subclass instance based on scorer_type
            instance = cls._create_metric_from_type(retrieved_scorer.scorer_type)
            StateManagementMixin.__init__(instance)
            instance._populate_from_scorer_response(retrieved_scorer)
            instance._set_state(SyncState.SYNCED)
            result.append(instance)

        return result

    @classmethod
    def delete_by_name(cls, name: str) -> None:
        """
        Delete a metric by name without retrieving it first.

        This is more efficient than calling `Metric.get(name=...).delete()`
        when you only need to delete and don't need the metric object.

        Args:
            name: The name of the metric to delete.

        Raises
        ------
            ValueError: If no metric with the given name exists.

        Examples
        --------
            # Delete without retrieving first
            Metric.delete_by_name("old-metric")

            # Alternative (less efficient)
            metric = Metric.get(name="old-metric")
            metric.delete()
        """
        logger.info(f"Metric.delete_by_name: name='{name}' - started")
        try:
            metrics_service = Metrics()
            metrics_service.delete_metric(name=name)
            logger.info(f"Metric.delete_by_name: name='{name}' - completed")
        except Exception as e:
            logger.error(f"Metric.delete_by_name: name='{name}' - failed: {e}")
            raise

    def _populate_from_scorer_response(self, scorer_response: Any) -> None:
        """Populate instance attributes from a ScorerResponse object."""
        # Common attributes for all metrics
        self.id = scorer_response.id
        self.name = scorer_response.name
        self.scorer_type = scorer_response.scorer_type
        self.tags = scorer_response.tags
        self.version = None

        # Handle optional common attributes
        self.description = (
            ""
            if isinstance(scorer_response.description, Unset) or scorer_response.description is None
            else scorer_response.description
        )
        self.created_at = None if isinstance(scorer_response.created_at, Unset) else scorer_response.created_at
        self.updated_at = None if isinstance(scorer_response.updated_at, Unset) else scorer_response.updated_at

        # Extract defaults - available for LLM and built-in Galileo metrics
        # These are returned by the API for preset scorers too
        if not isinstance(scorer_response.defaults, Unset) and scorer_response.defaults is not None:
            self.model = (
                scorer_response.defaults.model_name if hasattr(scorer_response.defaults, "model_name") else None
            )
            self.judges = (
                scorer_response.defaults.num_judges if hasattr(scorer_response.defaults, "num_judges") else None
            )
            self.cot_enabled = (
                scorer_response.defaults.cot_enabled if hasattr(scorer_response.defaults, "cot_enabled") else None
            )
        else:
            self.model = None
            self.judges = None
            self.cot_enabled = None

        # LLM-specific attributes (only set if this is an LlmMetric)
        if isinstance(self, LlmMetric):
            self.output_type = None if isinstance(scorer_response.output_type, Unset) else scorer_response.output_type
            self.prompt = None if isinstance(scorer_response.user_prompt, Unset) else scorer_response.user_prompt

            # Extract scoreable node types
            if not isinstance(scorer_response.scoreable_node_types, Unset) and scorer_response.scoreable_node_types:
                try:
                    self.node_level = StepType(scorer_response.scoreable_node_types[0])
                except (ValueError, IndexError):
                    self.node_level = None
            else:
                self.node_level = None

        # Code-specific attributes (only set if this is a CodeMetric)
        if isinstance(self, CodeMetric):
            # Extract scoreable node types
            if not isinstance(scorer_response.scoreable_node_types, Unset) and scorer_response.scoreable_node_types:
                try:
                    self.node_level = StepType(scorer_response.scoreable_node_types[0])
                except (ValueError, IndexError):
                    self.node_level = None
            else:
                self.node_level = None

    def update(self, **kwargs: Any) -> None:
        """
        Update this metric's properties.

        Currently not implemented as the API doesn't support updating scorers.

        Raises
        ------
            NotImplementedError: Always raised as updates are not supported.
        """
        raise NotImplementedError(
            "Updating metrics is not yet supported by the API. "
            "Consider creating a new metric with the desired properties instead."
        )

    def delete(self) -> None:
        """
        Delete this metric.

        Only works for server-side metrics. Local metrics don't need deletion.

        Raises
        ------
            ValidationError: If this is a local metric.
            ValueError: If the metric is not synced.

        Examples
        --------
            metric = Metric.get(name="factuality-checker")
            metric.delete()
        """
        if isinstance(self, LocalMetric):
            raise ValidationError("Local metrics don't exist on the server and can't be deleted.")

        if self.id is None:
            raise ValueError("Metric ID is not set. Cannot delete a local-only metric.")

        try:
            logger.info(f"Metric.delete: id='{self.id}' name='{self.name}' - started")
            metrics_service = Metrics()
            metrics_service.delete_metric(name=self.name)
            self._set_state(SyncState.DELETED)
            logger.info(f"Metric.delete: id='{self.id}' - completed")
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"Metric.delete: id='{self.id}' - failed: {e}")
            raise

    def refresh(self) -> None:
        """
        Refresh this metric's state from the API.

        Updates all attributes with the latest values from the remote API.

        Raises
        ------
            ValidationError: If this is a local metric.
            ValueError: If the metric is not synced.
            Exception: If the API call fails or the metric no longer exists.

        Examples
        --------
            metric.refresh()
            assert metric.is_synced()
        """
        if isinstance(self, LocalMetric):
            raise ValidationError("Local metrics don't exist on the server and can't be refreshed.")

        if self.id is None:
            raise ValueError("Metric ID is not set. Cannot refresh a local-only metric.")

        try:
            logger.debug(f"Metric.refresh: id='{self.id}' - started")
            scorers_service = Scorers()
            scorers = scorers_service.list()
            retrieved_scorer = next((s for s in scorers if s.id == self.id), None)

            if retrieved_scorer is None:
                raise ValueError(f"Metric with id '{self.id}' no longer exists")

            self._populate_from_scorer_response(retrieved_scorer)
            self._set_state(SyncState.SYNCED)
            logger.debug(f"Metric.refresh: id='{self.id}' - completed")
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"Metric.refresh: id='{self.id}' - failed: {e}")
            raise

    def to_legacy_metric(self) -> LegacyMetric:
        """
        Convert to legacy galileo.schema.metrics.Metric format.

        This enables backward compatibility with existing code that uses
        the legacy Metric class.

        Returns
        -------
            LegacyMetric: Legacy metric object with name and version.

        Examples
        --------
            metric = Metric.get(name="my-metric")
            legacy = metric.to_legacy_metric()
            # Use with existing APIs
        """
        return LegacyMetric(name=self.name, version=self.version)

    def __str__(self) -> str:
        """String representation of the metric."""
        type_name = self.__class__.__name__
        scorer_type_str = self.scorer_type.value if self.scorer_type else "unknown"
        return f"{type_name}(name='{self.name}', id='{self.id}', scorer_type='{scorer_type_str}')"

    def __repr__(self) -> str:
        """Detailed string representation of the metric."""
        type_name = self.__class__.__name__
        return f"{type_name}(name='{self.name}', id='{self.id}')"


# ============================================================================
# Concrete Metric Types
# ============================================================================


class LlmMetric(Metric):
    """
    LLM-based metric with custom prompt templates.

    This metric type allows you to create custom metrics evaluated by an LLM
    judge using a prompt template.

    Attributes
    ----------
        prompt (str | None): Prompt template for the LLM scorer.
        model (str | None): Model name/alias to use for scoring (stored as string).
        judges (int | None): Number of judges to use for scoring.
        cot_enabled (bool | None): Whether chain-of-thought is enabled.
        node_level (StepType | None): Node level for the metric.
        output_type (OutputTypeEnum | None): Output type for the metric.

    Configuration
    -------------
        Default values for `model` and `judges` can be configured via:
        - Configuration.default_scorer_model (env: GALILEO_DEFAULT_SCORER_MODEL)
        - Configuration.default_scorer_judges (env: GALILEO_DEFAULT_SCORER_JUDGES)

    Examples
    --------
        # Create custom LLM metric with string model name
        metric = LlmMetric(
            name="response_quality",
            prompt='''
            Rate the quality of this response on a scale of 1-10.

            Question: {input}
            Answer: {output}

            Return only the numerical score (1-10).
            ''',
            model="gpt-4o-mini",  # String model name
            judges=3,
            node_level=StepType.llm,
            description="Rates response quality",
            tags=["quality", "custom"],
            output_type=OutputTypeEnum.PERCENTAGE,
            cot_enabled=True,
        ).create()

        # Or use a Model object from Integration
        from galileo.__future__ import Integration
        gpt_model = Integration.openai.get_model(alias="gpt-4o-mini")
        metric = LlmMetric(
            name="response_quality",
            prompt="Rate quality 1-10: {input} -> {output}",
            model=gpt_model,  # Model object
            judges=3,
        ).create()
    """

    # Type annotations for LLM-specific attributes
    prompt: str | None
    model: str | None
    judges: int | None
    cot_enabled: bool | None
    node_level: StepType | None
    output_type: OutputTypeEnum | None

    def __init__(
        self,
        name: str,
        *,
        # LLM metric parameters (improved API)
        prompt: str | None = None,
        model: Model | str | None = None,
        judges: int | None = None,
        # Backward compatibility aliases
        user_prompt: str | None = None,
        model_name: str | None = None,
        num_judges: int | None = None,
        # LLM-specific parameters
        node_level: StepType | None = None,
        cot_enabled: bool | None = None,
        output_type: str | OutputTypeEnum | None = None,
        # Common parameters
        description: str = "",
        tags: list[str] | None = None,
        version: int | None = None,
    ) -> None:
        """
        Initialize an LLM metric.

        Args:
            name: The name of the metric.
            prompt: Prompt template for LLM scorers (preferred over user_prompt).
            model: Model object or model name string to use (preferred over model_name).
                   Defaults to Configuration.default_scorer_model.
            judges: Number of judges (preferred over num_judges). Defaults to Configuration.default_scorer_judges.
            user_prompt: [Deprecated] Use 'prompt' instead.
            model_name: [Deprecated] Use 'model' instead.
            num_judges: [Deprecated] Use 'judges' instead.
            node_level: Node level for the metric. Defaults to StepType.llm.
            cot_enabled: Whether chain-of-thought is enabled. Defaults to True.
            output_type: Output type ("percentage", "boolean", etc.).
            description: Description of the metric.
            tags: Tags associated with the metric.
            version: Specific version to reference (for existing metrics).

        Raises
        ------
            ValidationError: If prompt is not provided.
        """
        super().__init__(name=name, description=description, tags=tags, version=version)

        # Handle parameter aliases (new names preferred)
        final_prompt = prompt or user_prompt

        # Handle model parameter - extract alias from Model object if needed
        if model is not None:
            # Local import to avoid circular dependency
            from galileo.__future__.model import Model

            final_model = model.alias if isinstance(model, Model) else model
        else:
            final_model = model_name or Configuration.default_scorer_model

        final_judges = (
            judges
            if judges is not None
            else (num_judges if num_judges is not None else Configuration.default_scorer_judges)
        )

        if final_prompt is None:
            raise ValidationError("'prompt' (or 'user_prompt') must be provided for LLM-based metrics.")

        # Initialize LLM-specific attributes
        self.prompt = final_prompt
        self.model = final_model  # Now always a string (alias)
        self.judges = final_judges
        self.node_level = node_level or StepType.llm
        self.cot_enabled = cot_enabled if cot_enabled is not None else True

        # Handle output_type (accept string or enum)
        if isinstance(output_type, str):
            # Map common string values to enum
            output_type_map = {
                "percentage": OutputTypeEnum.PERCENTAGE,
                "boolean": OutputTypeEnum.BOOLEAN,
                "categorical": OutputTypeEnum.CATEGORICAL,
                "count": OutputTypeEnum.COUNT,
                "discrete": OutputTypeEnum.DISCRETE,
                "freeform": OutputTypeEnum.FREEFORM,
                "multilabel": OutputTypeEnum.MULTILABEL,
            }
            self.output_type = output_type_map.get(output_type.lower(), OutputTypeEnum.PERCENTAGE)
        else:
            self.output_type = output_type or OutputTypeEnum.BOOLEAN

        self.scorer_type = ScorerTypes.LLM

    def create(self) -> LlmMetric:
        """
        Persist this LLM metric to the API.

        Returns
        -------
            LlmMetric: This metric instance with updated attributes from the API.

        Raises
        ------
            ValidationError: If configuration is invalid.
            Exception: If the API call fails.

        Examples
        --------
            metric = LlmMetric(
                name="quality_check",
                prompt="Rate the quality...",
                model="gpt-4o-mini"
            ).create()
            assert metric.is_synced()
        """
        try:
            logger.info(f"LlmMetric.create: name='{self.name}' - started")

            metrics_service = Metrics()
            created_version = metrics_service.create_custom_llm_metric(
                name=self.name,
                user_prompt=self.prompt or "",
                node_level=self.node_level if self.node_level is not None else StepType.llm,
                cot_enabled=self.cot_enabled if self.cot_enabled is not None else True,
                model_name=self.model if self.model is not None else Configuration.default_scorer_model,
                num_judges=self.judges if self.judges is not None else Configuration.default_scorer_judges,
                description=self.description,
                tags=self.tags,
                output_type=self.output_type
                if isinstance(self.output_type, OutputTypeEnum)
                else OutputTypeEnum.BOOLEAN,
            )

            # Update attributes from response
            self.id = str(created_version.scorer_id)
            self.created_at = created_version.created_at
            self.updated_at = created_version.updated_at

            # Refresh to get full scorer details
            self.refresh()

            logger.info(f"LlmMetric.create: id='{self.id}' - completed")
            return self
        except ValidationError:
            raise
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"LlmMetric.create: name='{self.name}' - failed: {e}")
            raise

    def __repr__(self) -> str:
        """Detailed string representation of the metric."""
        return f"LlmMetric(name='{self.name}', id='{self.id}', model='{self.model}', judges={self.judges})"


class CodeMetric(Metric):
    r"""
    Code-based metric.

    This metric type is for code-based scorers that execute custom code
    to evaluate traces/spans.

    Attributes
    ----------
        node_level (StepType | None): Node level for the metric.
        code (str | None): The Python code for the scorer.

    Examples
    --------
        # Get existing code metric
        metric = Metric.get(name="my-code-metric")
        assert isinstance(metric, CodeMetric)

        # Create code metric with inline code
        metric = CodeMetric(
            name="custom_code_scorer",
            code="def scorer_fn(step_object):\\n    return 1.0",
            description="Custom code-based scorer",
            tags=["custom", "code"],
            node_level=StepType.llm,
        ).create()

        # Load code from file
        metric = CodeMetric(
            name="custom_code_scorer",
            node_level=StepType.llm,
        ).load_code("./scorers/my_scorer.py").create()
    """

    # Type annotations for code-specific attributes
    node_level: StepType | None
    code: str | None
    required_metrics: list[str] | None

    def __init__(
        self,
        name: str,
        *,
        code: str | None = None,
        node_level: StepType | None = None,
        required_metrics: list[str] | None = None,
        description: str = "",
        tags: list[str] | None = None,
        version: int | None = None,
    ) -> None:
        """
        Initialize a Code metric.

        Args:
            name: The name of the metric.
            code: The Python code for the scorer (optional, can be set later or loaded from file).
            node_level: Node level for the metric. Defaults to StepType.llm.
            required_metrics: List of metric names that this code metric depends on.
            description: Description of the metric.
            tags: Tags associated with the metric.
            version: Specific version to reference (for existing metrics).
        """
        super().__init__(name=name, description=description, tags=tags, version=version)

        self.code = code
        self.node_level = node_level or StepType.llm
        self.required_metrics = required_metrics
        self.scorer_type = ScorerTypes.CODE

    def load_code(self, code_file_path: str) -> CodeMetric:
        """
        Load code from a file into this metric instance.

        Args:
            code_file_path: Path to the Python file containing the scorer code.

        Returns
        -------
            CodeMetric: This metric instance with code loaded from the file (for chaining).

        Raises
        ------
            ValidationError: If the code file doesn't exist or can't be read.

        Examples
        --------
            # Load code from file
            metric = CodeMetric(
                name="custom_code_scorer",
                node_level=StepType.llm,
            ).load_code("./scorers/my_scorer.py").create()
        """
        if not os.path.isfile(code_file_path):
            raise ValidationError(f"Code file not found: {code_file_path}")

        try:
            with open(code_file_path, encoding="utf-8") as f:
                self.code = f.read()
        except Exception as e:
            raise ValidationError(f"Failed to read code file {code_file_path}: {e}")

        return self

    def _validate_code(self, config: GalileoPythonConfig) -> str:
        """
        Validate the code by submitting it to the validation endpoint and polling for results.

        Args:
            config: The Galileo configuration with API client.

        Returns
        -------
            str: The validation result as a JSON string to pass to create_code_scorer_version.

        Raises
        ------
            ValidationError: If validation fails or the code is invalid.
            ValueError: If the API returns an unexpected response.
        """
        assert self.code is not None
        assert self.node_level is not None

        # Step 1: Submit the code for validation
        code_bytes = self.code.encode("utf-8")
        code_file = File(payload=code_bytes, file_name="scorer.py")
        validate_body = BodyValidateCodeScorerScorersCodeValidatePost(
            file=code_file, scoreable_node_types=[self.node_level.value], required_scorers=self.required_metrics
        )

        validate_response = validate_code_scorer_scorers_code_validate_post.sync(
            client=config.api_client, body=validate_body
        )

        if validate_response is None:
            logger.debug("CodeMetric._validate_code: No response from validate_code_scorer")
            raise ValueError("Failed to validate code: No response from API")

        task_id = validate_response.task_id
        logger.debug(f"CodeMetric._validate_code: task_id='{task_id}' - validation started")

        # Step 2: Poll for validation result with time-based timeout
        start_time = time.time()
        attempt = 0
        while True:
            elapsed = time.time() - start_time
            timeout = Configuration.code_validation_timeout
            if elapsed >= timeout:
                raise ValidationError(f"Code validation timed out after {timeout:.0f} seconds")

            task_result = get_validate_code_scorer_task_result_scorers_code_validate_task_id_get.sync(
                task_id=task_id, client=config.api_client
            )

            if task_result is None:
                logger.debug(f"CodeMetric._validate_code: No response for task_id='{task_id}'")
                raise ValueError("Failed to get validation result: No response from API")

            if task_result.status == TaskResultStatus.COMPLETED:
                logger.debug(f"CodeMetric._validate_code: task_id='{task_id}' - validation completed")

                # Extract and validate the result
                result = task_result.result

                # Handle string result (already serialized)
                if isinstance(result, str):
                    return result

                # Handle ValidateRegisteredScorerResult or similar objects with to_dict
                if hasattr(result, "to_dict"):
                    # Check if it's an invalid result (has error_message in nested result)
                    if hasattr(result, "result") and isinstance(result.result, InvalidResult):
                        raise ValidationError(f"Code validation failed: {result.result.error_message}")
                    # Return the result as JSON string
                    return json.dumps(result.to_dict())

                raise ValueError(f"Unexpected validation result type: {type(result)}")

            if task_result.status == TaskResultStatus.FAILED:
                error_msg = "Code validation failed"
                if isinstance(task_result.result, str):
                    error_msg = f"Code validation failed: {task_result.result}"
                raise ValidationError(error_msg)

            if task_result.status == TaskResultStatus.PENDING:
                # Calculate delay with exponential backoff
                delay = min(
                    Configuration.code_validation_initial_delay
                    * (Configuration.code_validation_backoff_multiplier**attempt),
                    Configuration.code_validation_max_delay,
                )

                logger.debug(
                    f"CodeMetric._validate_code: task_id='{task_id}' - pending "
                    f"(elapsed: {elapsed:.1f}s/{timeout:.0f}s, next delay: {delay:.2f}s)"
                )
                time.sleep(delay)
                attempt += 1
            else:
                raise ValueError(f"Unknown task status: {task_result.status}")

    def create(self) -> CodeMetric:
        r"""
        Persist this Code metric to the API.

        This method validates the code first by submitting it to the validation
        endpoint, polling for the result, and then creating the scorer with the
        validated result.

        Returns
        -------
            CodeMetric: This metric instance with updated attributes from the API.

        Raises
        ------
            ValidationError: If code is not set, validation fails, or configuration is invalid.
            Exception: If the API call fails.

        Examples
        --------
            # Create with inline code
            metric = CodeMetric(
                name="custom_code_scorer",
                code="def scorer_fn(step_object):\\n    return 1.0",
                node_level=StepType.llm,
            ).create()
            assert metric.is_synced()

            # Create by loading from file
            metric = CodeMetric(
                name="custom_code_scorer",
                node_level=StepType.llm,
            ).load_code("./scorers/my_scorer.py").create()
            assert metric.is_synced()
        """
        # Validate that code is set
        if self.code is None:
            raise ValidationError(
                "Code is not set. Either pass 'code' to __init__() or use CodeMetric.load_code() to load from a file."
            )

        try:
            logger.info(f"CodeMetric.create: name='{self.name}' - started")

            config = GalileoPythonConfig.get()

            # Ensure node_level is set (should always be set in __init__, but checking for type safety)
            assert self.node_level is not None

            # Step 1: Validate the code and get validation result
            logger.debug(f"CodeMetric.create: name='{self.name}' - validating code")
            validation_result = self._validate_code(config)
            logger.debug(f"CodeMetric.create: name='{self.name}' - code validated successfully")

            # Step 2: Create the scorer
            scorer_request = CreateScorerRequest(
                name=self.name,
                scorer_type=ScorerTypes.CODE,
                description=self.description,
                tags=self.tags,
                scoreable_node_types=[self.node_level.value],
                required_scorers=self.required_metrics,
            )

            scorer_response = create_scorers_post.sync(client=config.api_client, body=scorer_request)

            if scorer_response is None:
                logger.debug("CodeMetric.create: No response from create_scorers_post")
                raise ValueError("Failed to create code-based metric: No response from API")

            # Step 3: Create the code scorer version with file upload and validation result
            # Convert the code string to bytes for file upload
            code_bytes = self.code.encode("utf-8")
            code_file = File(payload=code_bytes, file_name="scorer.py")
            version_body = BodyCreateCodeScorerVersionScorersScorerIdVersionCodePost(
                file=code_file, validation_result=validation_result
            )

            created_version = create_code_scorer_version_scorers_scorer_id_version_code_post.sync(
                scorer_id=scorer_response.id, client=config.api_client, body=version_body
            )

            if created_version is None:
                logger.debug(
                    "CodeMetric.create: No response from create_code_scorer_version_scorers_scorer_id_version_code_post"
                )
                raise ValueError("Failed to create code-based metric: No response from API")

            # Update attributes from response
            self.id = str(scorer_response.id)
            self.created_at = scorer_response.created_at
            self.updated_at = scorer_response.updated_at

            # Refresh to get full scorer details
            self.refresh()

            logger.info(f"CodeMetric.create: id='{self.id}' - completed")
            return self
        except ValidationError:
            raise
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"CodeMetric.create: name='{self.name}' - failed: {e}")
            raise

    def __repr__(self) -> str:
        """Detailed string representation of the metric."""
        return f"CodeMetric(name='{self.name}', id='{self.id}'')"


class GalileoMetric(Metric):
    """
    Built-in Galileo scorer metric.

    This metric type represents Galileo's built-in scorers like correctness,
    completeness, toxicity, etc. Access these via `Metric.metrics`.

    Examples
    --------
        # Access built-in scorers
        from galileo.__future__ import Metric, LogStream

        log_stream = LogStream.get(name="my-stream", project_name="my-project")
        log_stream.set_metrics([
            Metric.metrics.correctness,
            Metric.metrics.completeness,
            Metric.metrics.toxicity,
        ])

        # Or get by name
        metric = Metric.get(name="correctness")
        assert isinstance(metric, GalileoMetric)
    """

    def __init__(
        self, name: str, *, description: str = "", tags: list[str] | None = None, version: int | None = None
    ) -> None:
        """
        Initialize a Galileo metric.

        Args:
            name: The name of the metric.
            description: Description of the metric.
            tags: Tags associated with the metric.
            version: Specific version to reference (for existing metrics).
        """
        super().__init__(name=name, description=description, tags=tags, version=version)
        # Galileo metrics can have various scorer types, set during population


class LocalMetric(Metric):
    """
    Local function-based metric.

    This metric type uses a Python function to score traces/spans locally
    without making API calls. Useful for simple, deterministic metrics.

    Attributes
    ----------
        scorer_fn (Callable): Scoring function that takes a Trace or Span and returns a score.
        scorable_types (list[StepType]): Types that can be scored.
        aggregatable_types (list[StepType]): Types that can be aggregated.

    Examples
    --------
        # Create local function-based metric
        def response_length_scorer(trace_or_span):
            if hasattr(trace_or_span, "output") and trace_or_span.output:
                return min(len(trace_or_span.output) / 100.0, 1.0)
            return 0.0

        local_metric = LocalMetric(
            name="response_length",
            scorer_fn=response_length_scorer,
            scorable_types=[StepType.llm],
            aggregatable_types=[StepType.trace],
        )

        # Use with log stream
        log_stream.set_metrics([local_metric])
    """

    # Type annotations for local metric attributes
    scorer_fn: Callable[[Trace | Span], MetricValueType]
    scorable_types: list[StepType]
    aggregatable_types: list[StepType]

    def __init__(
        self,
        name: str,
        *,
        scorer_fn: Callable[[Trace | Span], MetricValueType],
        scorable_types: list[StepType] | None = None,
        aggregatable_types: list[StepType] | None = None,
        description: str = "",
        tags: list[str] | None = None,
    ) -> None:
        """
        Initialize a local function-based metric.

        Args:
            name: The name of the metric.
            scorer_fn: Scoring function for the metric.
            scorable_types: Step types that can be scored. Defaults to [StepType.llm].
            aggregatable_types: Step types for aggregation. Defaults to [StepType.trace].
            description: Description of the metric.
            tags: Tags associated with the metric.

        Raises
        ------
            ValidationError: If scorer_fn is not provided.
        """
        super().__init__(name=name, description=description, tags=tags)

        if scorer_fn is None:
            raise ValidationError("'scorer_fn' must be provided for local metrics.")

        self.scorer_fn = scorer_fn
        self.scorable_types = scorable_types or [StepType.llm]
        self.aggregatable_types = aggregatable_types or [StepType.trace]
        self.scorer_type = None  # Local metrics don't have a scorer_type

    def to_local_metric_config(self) -> LocalMetricConfig:
        """
        Convert to LocalMetricConfig format.

        Returns
        -------
            LocalMetricConfig: Local metric configuration for use with the logger.

        Examples
        --------
            def my_scorer(trace):
                return 0.5

            metric = LocalMetric(name="test", scorer_fn=my_scorer)
            config = metric.to_local_metric_config()
        """
        return LocalMetricConfig(
            name=self.name,
            scorer_fn=self.scorer_fn,
            scorable_types=self.scorable_types,
            aggregatable_types=self.aggregatable_types,
        )

    def __repr__(self) -> str:
        """Detailed string representation of the metric."""
        # Handle callables that don't have __name__ (partials, lambdas, callable instances)
        fn_name = getattr(self.scorer_fn, "__name__", f"<{type(self.scorer_fn).__name__}>")
        return f"LocalMetric(name='{self.name}', scorer_fn={fn_name})"



from enum import Enum


class InputTypeEnum(str, Enum):
    AGENT_SPANS = "agent_spans"
    BASIC = "basic"
    LLM_SPANS = "llm_spans"
    RETRIEVER_SPANS = "retriever_spans"
    SESSIONS_NORMALIZED = "sessions_normalized"
    SESSIONS_TRACE_IO_ONLY = "sessions_trace_io_only"
    TOOL_SPANS = "tool_spans"
    TRACE_INPUT_ONLY = "trace_input_only"
    TRACE_IO_ONLY = "trace_io_only"
    TRACE_NORMALIZED = "trace_normalized"
    TRACE_OUTPUT_ONLY = "trace_output_only"
    WORKFLOW_SPANS = "workflow_spans"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class CoreScorerName(str, Enum):
    ACTION_ADVANCEMENT_LUNA = "action_advancement_luna"
    ACTION_COMPLETION_LUNA = "action_completion_luna"
    AGENTIC_SESSION_SUCCESS = "agentic_session_success"
    AGENTIC_WORKFLOW_SUCCESS = "agentic_workflow_success"
    AGENT_EFFICIENCY = "agent_efficiency"
    AGENT_FLOW = "agent_flow"
    BLEU = "bleu"
    CHUNK_ATTRIBUTION_UTILIZATION = "chunk_attribution_utilization"
    CHUNK_ATTRIBUTION_UTILIZATION_LUNA = "chunk_attribution_utilization_luna"
    COMPLETENESS = "completeness"
    COMPLETENESS_LUNA = "completeness_luna"
    CONTEXT_ADHERENCE = "context_adherence"
    CONTEXT_ADHERENCE_LUNA = "context_adherence_luna"
    CONTEXT_RELEVANCE = "context_relevance"
    CONTEXT_RELEVANCE_LUNA = "context_relevance_luna"
    CONVERSATION_QUALITY = "conversation_quality"
    CORRECTNESS = "correctness"
    GROUND_TRUTH_ADHERENCE = "ground_truth_adherence"
    INPUT_PII = "input_pii"
    INPUT_PII_GPT = "input_pii_gpt"
    INPUT_SEXIST = "input_sexist"
    INPUT_SEXIST_LUNA = "input_sexist_luna"
    INPUT_TONE = "input_tone"
    INPUT_TONE_GPT = "input_tone_gpt"
    INPUT_TOXICITY = "input_toxicity"
    INPUT_TOXICITY_LUNA = "input_toxicity_luna"
    INSTRUCTION_ADHERENCE = "instruction_adherence"
    OUTPUT_PII = "output_pii"
    OUTPUT_PII_GPT = "output_pii_gpt"
    OUTPUT_SEXIST = "output_sexist"
    OUTPUT_SEXIST_LUNA = "output_sexist_luna"
    OUTPUT_TONE = "output_tone"
    OUTPUT_TONE_GPT = "output_tone_gpt"
    OUTPUT_TOXICITY = "output_toxicity"
    OUTPUT_TOXICITY_LUNA = "output_toxicity_luna"
    PROMPT_INJECTION = "prompt_injection"
    PROMPT_INJECTION_LUNA = "prompt_injection_luna"
    PROMPT_PERPLEXITY = "prompt_perplexity"
    ROUGE = "rouge"
    SQL_ADHERENCE = "sql_adherence"
    SQL_CORRECTNESS = "sql_correctness"
    SQL_EFFICIENCY = "sql_efficiency"
    SQL_INJECTION = "sql_injection"
    TOOL_ERROR_RATE = "tool_error_rate"
    TOOL_ERROR_RATE_LUNA = "tool_error_rate_luna"
    TOOL_SELECTION_QUALITY = "tool_selection_quality"
    TOOL_SELECTION_QUALITY_LUNA = "tool_selection_quality_luna"
    UNCERTAINTY = "uncertainty"
    USER_INTENT_CHANGE = "user_intent_change"

    def __str__(self) -> str:
        return str(self.value)
